# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Point3d:
    """Dimensional information of an object in 3d.

    Parameters
    ----------
    x: float
        Position of the object in the x-dimension.
    y: float
        Position of the object in the y-dimension.
    z: float
        Position of the object in the z-dimension.
    """

    x: float
    y: float
    z: float

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Point3d":
        """Generate a Point3d from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Point3d
            Converted 3d point.
        """

        return Point3d(
            x=float(data_dict["x"]),
            y=float(data_dict["y"]),
            z=float(data_dict["z"]),
        )



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Point2d:
    """A 2d point in an image.

    Parameters
    ----------
    x: float or int
        The x-coordinate of the point in the image.
    y: float or int
        The y-coordinate of the point in the image.
    """

    x: float
    y: float

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return [float(self.x), float(self.y)]



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from decimal import Decimal

from ._filter_abc import Frame, _FilterABC


class _FilterEnd(_FilterABC):

    PARAMETERS = ["end_frame", "end_timestamp"]
    LEVELS = ["frame"]

    def passes_filter(self, frame: Frame) -> bool:

        if self.end_frame is not None:
            return frame.uid <= self.end_frame

        elif self.end_timestamp is not None:
            return frame.timestamp <= Decimal(self.end_timestamp)

        else:
            return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
import uuid
from dataclasses import dataclass, field

from .frame import Frame
from .metadata import Metadata
from .object import Object
from .sensor import Sensor


@dataclass
class Scene:
    """The root RailLabel class, which contains all data.

    Parameters
    ----------
    metadata: raillabel.format.Metadata
        This object contains information, that is, metadata, about the annotation file itself.
    sensors: dict of raillabel.format.Sensor, optional
        Dictionary of raillabel.format.Sensors. Dictionary keys are the sensor uids. Default is {}.
    objects: dict of raillabel.format.Object, optional
        Dictionary of raillabel.format.Objects. Dictionary keys are the object uids. Default is {}.
    frames: dict of raillabel.format.Frame, optional
        Dict of frames in the scene. Dictionary keys are the frame uids. Default is {}.
    """

    metadata: Metadata
    sensors: t.Dict[str, Sensor] = field(default_factory=dict)
    objects: t.Dict[uuid.UUID, Object] = field(default_factory=dict)
    frames: t.Dict[int, Frame] = field(default_factory=dict)

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this Scene.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {"openlabel": {"metadata": self.metadata.asdict()}}

        if self.sensors != {}:
            dict_repr["openlabel"]["streams"] = {
                str(k): v.asdict()["stream"] for k, v in self.sensors.items()
            }
            dict_repr["openlabel"]["coordinate_systems"] = {
                str(k): v.asdict()["coordinate_system"] for k, v in self.sensors.items()
            }
            dict_repr["openlabel"]["coordinate_systems"]["base"] = {
                "type": "local",
                "parent": "",
                "children": list(self.sensors.keys()),
            }

        if self.objects != {}:
            dict_repr["openlabel"]["objects"] = {
                str(k): v.asdict() for k, v in self.objects.items()
            }

        if self.frames != {}:
            dict_repr["openlabel"]["frames"] = {str(k): v.asdict() for k, v in self.frames.items()}

        dict_repr = self._add_object_data_pointers(dict_repr)

        return dict_repr

    def _add_object_data_pointers(self, dict_repr: dict) -> dict:
        """Add object frame intervals and object_data_pointers to the JSON.

        The VCD visualization tool requires a short "summary" of where an object and the
        object_data_pointers occurr, which includes the frame intervals of the object and the frame
        intervals of the object_data names.

        Parameters
        ----------
        dict_repr : dict
            Dictionary in the RailLabel format, that should be enhanced.

        Returns
        -------
        dict
            Enhanced dictionary.
        """

        # Creates the frame intervals and object_data_pointers
        frame_intervals = []
        object_frame_intervals = {}
        for frame in self.frames.values():

            # Adds the frame intervals
            if len(frame_intervals) == 0:
                frame_intervals.append(
                    {
                        "frame_start": frame.uid,
                        "frame_end": frame.uid,
                    }
                )

            elif frame.uid == frame_intervals[-1]["frame_end"] + 1:
                frame_intervals[-1]["frame_end"] += 1

            else:
                frame_intervals.append(
                    {
                        "frame_start": frame.uid,
                        "frame_end": frame.uid,
                    }
                )

            # Adds the object frame intervals
            for object_id in frame.object_data:

                if object_id not in object_frame_intervals:
                    object_frame_intervals[object_id] = {
                        "frame_intervals": [
                            {
                                "frame_start": frame.uid,
                                "frame_end": frame.uid,
                            }
                        ],
                        "object_data_pointers": {},
                    }

                else:
                    if (
                        object_frame_intervals[object_id]["frame_intervals"][-1]["frame_end"]
                        == frame.uid - 1
                    ):
                        object_frame_intervals[object_id]["frame_intervals"][-1][
                            "frame_end"
                        ] = frame.uid

                    else:
                        object_frame_intervals[object_id]["frame_intervals"].append(
                            {
                                "frame_start": frame.uid,
                                "frame_end": frame.uid,
                            }
                        )

            # Adds the object_data_pointers
            for object_id in frame.object_data:
                for annotation in frame.object_data[object_id].annotations.values():

                    if (
                        annotation.name
                        not in object_frame_intervals[object_id]["object_data_pointers"]
                    ):
                        object_frame_intervals[object_id]["object_data_pointers"][
                            annotation.name
                        ] = {
                            "type": annotation.name.split("__")[1],
                            "frame_intervals": [
                                {
                                    "frame_start": frame.uid,
                                    "frame_end": frame.uid,
                                }
                            ],
                            "attribute_pointers": {},
                        }

                    else:
                        if (
                            object_frame_intervals[object_id]["object_data_pointers"][
                                annotation.name
                            ]["frame_intervals"][-1]["frame_end"]
                            >= frame.uid - 1
                        ):
                            object_frame_intervals[object_id]["object_data_pointers"][
                                annotation.name
                            ]["frame_intervals"][-1]["frame_end"] = frame.uid

                        else:
                            object_frame_intervals[object_id]["object_data_pointers"][
                                annotation.name
                            ]["frame_intervals"].append(
                                {
                                    "frame_start": frame.uid,
                                    "frame_end": frame.uid,
                                }
                            )

                    for attr_name, attr_value in annotation.attributes.items():

                        if type(attr_value) == str:
                            attr_type = "text"

                        elif type(attr_value) in [float, int]:
                            attr_type = "num"

                        elif type(attr_value) == bool:
                            attr_type = "boolean"

                        elif type(attr_value) in [list, tuple]:
                            attr_type = "vec"

                        object_frame_intervals[object_id]["object_data_pointers"][annotation.name][
                            "attribute_pointers"
                        ][attr_name] = attr_type

        # Adds the frame_intervals to the dict
        dict_repr["openlabel"]["frame_intervals"] = frame_intervals

        # Adds the object_data_pointers to the dict
        if "objects" in dict_repr["openlabel"]:
            for object_id, object in dict_repr["openlabel"]["objects"].items():
                object.update(object_frame_intervals[object_id])

        return dict_repr



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from uuid import UUID

from ._annotation import _Annotation
from .sensor_reference import SensorReference


@dataclass
class Polyline2d(_Annotation):
    """A 2d polyline.

    Parameters
    ----------
    id: uuid.UUID
        Unique identifier of the annotation.
    object_id: uuid.UUID
        Unique identifier of the object this annotation refers to. Used for tracking.
    class_name: str
        Name of the class this annotation belongs to.
    attributes: dict[str, str or list]
        Key value pairs of attributes with the keys beeing the friendly identifier of the attribute and the value beeing the attribute value.
    sensor: raillabel._understand_ai_t4_format.SensorReference
        Information about the sensor this annotation is labeled in.
    points: list[tuple[float, float]]
        2d points belonging to the polyline.
    """

    points: t.List[t.Tuple[float, float]]

    OPENLABEL_ID = "poly2d"

    @classmethod
    def fromdict(cls, data_dict: t.Dict) -> "Polyline2d":
        """Generate a Polyline2d from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Polyline2d
            Converted 2d polyline.
        """

        return Polyline2d(
            id=UUID(data_dict["id"]),
            object_id=UUID(data_dict["objectId"]),
            class_name=data_dict["className"],
            attributes=data_dict["attributes"],
            sensor=SensorReference.fromdict(data_dict["sensor"]),
            points=[(p[0], p[1]) for p in data_dict["geometry"]["points"]],
        )

    def to_raillabel(self) -> t.Tuple[dict, str, str, dict]:
        """Convert to a raillabel compatible dict.

        Returns
        -------
        annotation: dict
            Dictionary valid for the raillabel schema.
        object_id: str
            Friendly identifier of the object this sensor belongs to.
        class_name: str
            Friendly identifier of the class the annotated object belongs to.
        sensor_reference: dict
            Dictionary of the sensor reference.
        """

        polyline = super().to_raillabel()
        polyline[0]["closed"] = False
        polyline[0]["mode"] = "MODE_POLY2D_ABSOLUTE"

        return polyline

    def _val_to_raillabel(self) -> t.List[float]:
        return [coordinate for point in self.points for coordinate in point]



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

from test_frame import _prepare_frame_data

import raillabel._understand_ai_t4_format as uai_format


def remove_generated_fields(raillabel_data: dict) -> dict:
    """Return RailLabel file with object_data_pointers and frame_intervals
    removed."""

    raillabel_data = _remove_frame_intervals(raillabel_data)
    raillabel_data = _remove_object_data_pointers(raillabel_data)
    raillabel_data = _remove_annotation_names(raillabel_data)

    return raillabel_data

def _remove_frame_intervals(raillabel_data: dict) -> dict:
    if "frame_intervals" in raillabel_data["openlabel"]:
        del raillabel_data["openlabel"]["frame_intervals"]

    for object in raillabel_data["openlabel"]["objects"]:
        if "frame_intervals" in raillabel_data["openlabel"]["objects"][object]:
            del raillabel_data["openlabel"]["objects"][object]["frame_intervals"]

    return raillabel_data

def _remove_object_data_pointers(raillabel_data: dict) -> dict:
    for object in raillabel_data["openlabel"]["objects"]:
        if "object_data_pointers" in raillabel_data["openlabel"]["objects"][object]:
            del raillabel_data["openlabel"]["objects"][object]["object_data_pointers"]

    return raillabel_data

def _remove_annotation_names(raillabel_data: dict) -> dict:
    for frame in raillabel_data["openlabel"]["frames"].values():
        for object in frame["objects"].values():
            for annotations in object["object_data"].values():
                for annotation in annotations:
                    if "name" in annotation:
                        annotation["name"] = annotation["uid"]
                        del annotation["uid"]

    return raillabel_data


def remove_non_parsed_fields(raillabel_data: dict) -> dict:
    """Return RailLabel file with frame_data and poly3ds removed."""

    for frame in raillabel_data["openlabel"]["frames"].values():

        if "frame_data" in frame["frame_properties"]:
            del frame["frame_properties"]["frame_data"]

        for object_id, object in list(frame["objects"].items()):
            if "poly3d" not in object["object_data"]:
                continue

            del object["object_data"]["poly3d"]
            if len(object["object_data"]) == 0:
                del frame["objects"][object_id]

    return raillabel_data


def test_fromdict(json_data):
    input_data = {
        "metadata": json_data["_understand_ai_t4_format/metadata"],
        "coordinateSystems": [
            json_data["_understand_ai_t4_format/coordinate_system_camera"],
            json_data["_understand_ai_t4_format/coordinate_system_lidar"],
        ],
        "frames": [
            _prepare_frame_data(json_data["_understand_ai_t4_format/frame"], json_data),
        ]
    }
    scene = uai_format.Scene.fromdict(input_data)

    assert scene.metadata == uai_format.Metadata.fromdict(input_data["metadata"])
    assert len(scene.coordinate_systems) == 2
    assert scene.coordinate_systems[input_data["coordinateSystems"][0]["coordinate_system_id"]] == uai_format.CoordinateSystem.fromdict(input_data["coordinateSystems"][0])
    assert scene.coordinate_systems[input_data["coordinateSystems"][1]["coordinate_system_id"]] == uai_format.CoordinateSystem.fromdict(input_data["coordinateSystems"][1])
    assert len(scene.frames) == 1
    assert scene.frames[input_data["frames"][0]["frameId"]] == uai_format.Frame.fromdict(input_data["frames"][0])


def test_to_raillabel(json_data):
    input_data = json_data["understand_ai_t4_short"]
    input_data["metadata"] = json_data["_understand_ai_t4_format/metadata"]
    scene = uai_format.Scene.fromdict(json_data["understand_ai_t4_short"])
    output_data = scene.to_raillabel()

    ground_truth = remove_generated_fields(json_data["openlabel_v1_short"])
    ground_truth = remove_non_parsed_fields(ground_truth)

    ground_truth["openlabel"]["metadata"] = None
    output_data["openlabel"]["metadata"] = None

    assert output_data == ground_truth


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from decimal import Decimal

from ._filter_abc import Frame, _FilterABC


class _FilterStart(_FilterABC):

    PARAMETERS = ["start_frame", "start_timestamp"]
    LEVELS = ["frame"]

    def passes_filter(self, frame: Frame) -> bool:

        if self.start_frame is not None:
            return frame.uid >= self.start_frame

        elif self.start_timestamp is not None:
            return frame.timestamp >= Decimal(self.start_timestamp)

        else:
            return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from decimal import Decimal
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict(json_data):
    input_data = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    sensor_reference = uai_format.SensorReference.fromdict(input_data)

    assert sensor_reference.type == input_data["type"]
    assert sensor_reference.uri == input_data["uri"]
    assert sensor_reference.timestamp == Decimal(input_data["timestamp"])

def test_to_raillabel(json_data):
    sensor_reference = uai_format.SensorReference.fromdict(
        json_data["_understand_ai_t4_format/sensor_reference_camera"]
    )
    sensor_id, output_data = sensor_reference.to_raillabel()

    assert sensor_id == sensor_reference.type
    assert output_data == {
        "stream_properties": {
            "sync": {
                "timestamp": str(sensor_reference.timestamp)
            }
        },
        "uri": sensor_reference.uri
    }


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Devkit for working with recorded and annotated train ride data from DB."""
from importlib import metadata

from . import format
from .exceptions import *
from .filter import filter
from .format.scene import Scene
from .load import load
from .save import save
from .validate import validate

try:
    __version__ = metadata.version("raillabel")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0+unknown"
del metadata



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import pathlib

import pytest

# Executes the test
os.system("clear")
pytest.main(
    [
        str(pathlib.Path(__file__).parent),
        "--disable-pytest-warnings",
        "--cache-clear",
    ]
)



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import glob
import json
import sys
import typing as t
from pathlib import Path

import json5
import pytest

sys.path.insert(1, str(Path(__file__).parent.parent))

import raillabel

json_data_directories = [
    Path(__file__).parent / "test_raillabel" / "__test_assets__",
    Path(__file__).parent.parent / "raillabel" / "schemas"
]

@pytest.fixture(scope="session", autouse=True)
def compile_uncommented_test_file():
    """Compiles the main test file from json5 to json."""

    json5_file_path = json_data_directories[0] / "openlabel_v1_short.json5"

    with json5_file_path.open() as f:
        data = json5.load(f)

    with open(str(json5_file_path)[:-1], "w") as f:
        json.dump(data, f, indent=4)


@pytest.fixture
def json_paths(request) -> t.Dict[str, Path]:
    json_paths = _fetch_json_paths_from_cache(request)

    if json_paths is None:
        json_paths = {_get_file_identifier(p): p for p in _collect_json_paths()}

    return json_paths

def _fetch_json_paths_from_cache(request) -> t.Optional[t.Dict[str, Path]]:
    return request.config.cache.get("json_paths", None)

def _collect_json_paths() -> t.List[Path]:
    json_paths = []

    for dir in json_data_directories:
        json_paths.extend([Path(p) for p in glob.glob(str(dir) + "/**/**.json", recursive=True)])

    return json_paths

def _get_file_identifier(path: Path) -> str:
    """Return relative path from test asset dir as string."""

    if "__test_assets__" not in path.parts:
        return path.stem

    test_assets_dir_index = path.parts.index("__test_assets__")

    relative_path = ""
    for part in path.parts[test_assets_dir_index+1:-1]:
        relative_path += part + "/"

    relative_path += path.stem

    return relative_path

@pytest.fixture
def json_data(request) -> t.Dict[str, dict]:
    json_data = _fetch_json_data_from_cache(request)

    if json_data is None:
        json_data = {_get_file_identifier(p): _load_json_data(p) for p in _collect_json_paths()}

    return json_data

def _fetch_json_data_from_cache(request) -> t.Optional[t.Dict[str, Path]]:
    return request.config.cache.get("json_data", None)

def _load_json_data(path: Path) -> dict:
    with path.open() as f:
        json_data = json.load(f)
    return json_data


@pytest.fixture
def annotation_compare_methods() -> t.Dict[str, t.Callable]:
    methods = {}

    def compare_bbox(annotation: raillabel.format.Bbox, ground_truth: dict) -> bool:

        assert type(annotation) == raillabel.format.Bbox
        assert annotation.uid == ground_truth["uid"]
        assert annotation.name == ground_truth["name"]
        assert annotation.pos.x == ground_truth["val"][0]
        assert annotation.pos.y == ground_truth["val"][1]
        assert annotation.size.x == ground_truth["val"][2]
        assert annotation.size.y == ground_truth["val"][3]

        if "coordinate_system" in ground_truth:
            assert annotation.sensor.uid == ground_truth["coordinate_system"]

        if "attributes" in ground_truth:

            accumulative_attributes = []
            for attr_type in ground_truth["attributes"].values():
                accumulative_attributes.extend(attr_type)

            assert len(annotation.attributes) == len(accumulative_attributes)
            for attribute in accumulative_attributes:
                assert attribute["name"] in annotation.attributes
                assert annotation.attributes[attribute["name"]] == attribute["val"]

    methods["bbox"] = compare_bbox


    def compare_cuboid(annotation: raillabel.format.Cuboid, ground_truth: dict) -> bool:

        assert type(annotation) == raillabel.format.Cuboid
        assert annotation.uid == ground_truth["uid"]
        assert annotation.name == ground_truth["name"]

        cuboid_val = [
            annotation.pos.x,
            annotation.pos.y,
            annotation.pos.z,
            annotation.quat.x,
            annotation.quat.y,
            annotation.quat.z,
            annotation.quat.w,
            annotation.size.x,
            annotation.size.y,
            annotation.size.z,
        ]
        assert cuboid_val == ground_truth["val"]

        if "coordinate_system" in ground_truth:
            assert annotation.sensor.uid == ground_truth["coordinate_system"]

        if "attributes" in ground_truth:

            accumulative_attributes = []
            for attr_type in ground_truth["attributes"].values():
                accumulative_attributes.extend(attr_type)

            assert len(annotation.attributes) == len(accumulative_attributes)
            for attribute in accumulative_attributes:
                assert attribute["name"] in annotation.attributes
                assert annotation.attributes[attribute["name"]] == attribute["val"]

    methods["cuboid"] = compare_cuboid


    def compare_num(annotation: raillabel.format.Num, ground_truth: dict) -> bool:

        assert type(annotation) == raillabel.format.Num
        assert annotation.uid == ground_truth["uid"]
        assert annotation.name == ground_truth["name"]
        assert annotation.val == ground_truth["val"]

        if "coordinate_system" in ground_truth:
            assert annotation.sensor.uid == ground_truth["coordinate_system"]

        if "attributes" in ground_truth:

            accumulative_attributes = []
            for attr_type in ground_truth["attributes"].values():
                accumulative_attributes.extend(attr_type)

            assert len(annotation.attributes) == len(accumulative_attributes)
            for attribute in accumulative_attributes:
                assert attribute["name"] in annotation.attributes
                assert annotation.attributes[attribute["name"]] == attribute["val"]

    methods["num"] = compare_num


    def compare_poly2d(annotation: raillabel.format.Poly2d, ground_truth: dict) -> bool:

        assert type(annotation) == raillabel.format.Poly2d
        assert annotation.uid == ground_truth["uid"]
        assert annotation.name == ground_truth["name"]

        for i in range(0, len(ground_truth["val"]), 2):
            assert annotation.points[i // 2].x == ground_truth["val"][i]
            assert annotation.points[i // 2].y == ground_truth["val"][i + 1]

        if "closed" in ground_truth:
            assert annotation.closed == ground_truth["closed"]

        if "mode" in ground_truth:
            assert annotation.mode == ground_truth["mode"]

        if "coordinate_system" in ground_truth:
            assert annotation.sensor.uid == ground_truth["coordinate_system"]

        if "attributes" in ground_truth:

            accumulative_attributes = []
            for attr_type in ground_truth["attributes"].values():
                accumulative_attributes.extend(attr_type)

            assert len(annotation.attributes) == len(accumulative_attributes)
            for attribute in accumulative_attributes:
                assert attribute["name"] in annotation.attributes
                assert annotation.attributes[attribute["name"]] == attribute["val"]

    methods["poly2d"] = compare_poly2d


    def compare_poly3d(annotation: raillabel.format.Poly3d, ground_truth: dict) -> bool:

        assert type(annotation) == raillabel.format.Poly3d
        assert annotation.uid == ground_truth["uid"]
        assert annotation.name == ground_truth["name"]

        for i in range(0, len(ground_truth["val"]), 3):
            assert annotation.points[int(i / 3)].x == ground_truth["val"][i]
            assert annotation.points[int(i / 3)].y == ground_truth["val"][i + 1]
            assert annotation.points[int(i / 3)].z == ground_truth["val"][i + 2]

        if "closed" in ground_truth:
            assert annotation.closed == ground_truth["closed"]

        if "coordinate_system" in ground_truth:
            assert annotation.sensor.uid == ground_truth["coordinate_system"]

        if "attributes" in ground_truth:

            accumulative_attributes = []
            for attr_type in ground_truth["attributes"].values():
                accumulative_attributes.extend(attr_type)

            assert len(annotation.attributes) == len(accumulative_attributes)
            for attribute in accumulative_attributes:
                assert attribute["name"] in annotation.attributes
                assert annotation.attributes[attribute["name"]] == attribute["val"]

    methods["poly3d"] = compare_poly3d


    def compare_seg3d(annotation: raillabel.format.Seg3d, ground_truth: dict) -> bool:

        assert type(annotation) == raillabel.format.Seg3d
        assert annotation.uid == ground_truth["uid"]
        assert annotation.name == ground_truth["name"]
        assert annotation.point_ids == ground_truth["val"]

        if "coordinate_system" in ground_truth:
            assert annotation.sensor.uid == ground_truth["coordinate_system"]

        if "attributes" in ground_truth:

            accumulative_attributes = []
            for attr_type in ground_truth["attributes"].values():
                accumulative_attributes.extend(attr_type)

            assert len(annotation.attributes) == len(accumulative_attributes)
            for attribute in accumulative_attributes:
                assert attribute["name"] in annotation.attributes
                assert annotation.attributes[attribute["name"]] == attribute["val"]

    methods["vec"] = compare_seg3d

    return methods



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict_camera(json_data):
    input_data = json_data["_understand_ai_t4_format/coordinate_system_camera"]
    coordinate_system = uai_format.CoordinateSystem.fromdict(input_data)

    assert coordinate_system.uid == input_data["coordinate_system_id"]
    assert coordinate_system.topic == input_data["topic"]
    assert coordinate_system.frame_id == input_data["frame_id"]
    assert coordinate_system.position == input_data["position"]
    assert coordinate_system.rotation_quaternion == input_data["rotation_quaternion"]
    assert coordinate_system.rotation_matrix == input_data["rotation_matrix"]
    assert coordinate_system.angle_axis_rotation == input_data["angle_axis_rotation"]
    assert coordinate_system.homogeneous_transform == input_data["homogeneous_transform"]
    assert coordinate_system.measured_position == input_data["measured_position"]
    assert coordinate_system.camera_matrix == input_data["camera_matrix"]
    assert coordinate_system.dist_coeffs == input_data["dist_coeffs"]

def test_fromdict_lidar(json_data):
    input_data = json_data["_understand_ai_t4_format/coordinate_system_lidar"]
    coordinate_system = uai_format.CoordinateSystem.fromdict(input_data)

    assert coordinate_system.uid == input_data["coordinate_system_id"]
    assert coordinate_system.topic == input_data["topic"]
    assert coordinate_system.frame_id == input_data["frame_id"]
    assert coordinate_system.position == input_data["position"]
    assert coordinate_system.rotation_quaternion == input_data["rotation_quaternion"]
    assert coordinate_system.rotation_matrix == input_data["rotation_matrix"]
    assert coordinate_system.angle_axis_rotation == input_data["angle_axis_rotation"]
    assert coordinate_system.homogeneous_transform == input_data["homogeneous_transform"]
    assert coordinate_system.measured_position == None
    assert coordinate_system.camera_matrix == None
    assert coordinate_system.dist_coeffs == None

def test_to_raillabel_camera(json_data):
    input_data = json_data["_understand_ai_t4_format/coordinate_system_camera"]
    coordinate_system = uai_format.CoordinateSystem.fromdict(input_data)
    output_cs, output_stream = coordinate_system.to_raillabel()
    ground_truth_cs = json_data["_understand_ai_t4_format/coordinate_system_camera_cs_raillabel"]
    ground_truth_stream = json_data["_understand_ai_t4_format/coordinate_system_camera_stream_raillabel"]

    assert output_cs == ground_truth_cs
    assert output_stream == ground_truth_stream

def test_to_raillabel_lidar(json_data):
    input_data = json_data["_understand_ai_t4_format/coordinate_system_lidar"]
    coordinate_system = uai_format.CoordinateSystem.fromdict(input_data)
    _, output_stream = coordinate_system.to_raillabel()
    ground_truth_stream = json_data["_understand_ai_t4_format/coordinate_system_lidar_stream_raillabel"]

    assert output_stream == ground_truth_stream

def test_to_raillabel_radar(json_data):
    input_data = json_data["_understand_ai_t4_format/coordinate_system_radar"]
    coordinate_system = uai_format.CoordinateSystem.fromdict(input_data)
    _, output_stream = coordinate_system.to_raillabel()
    ground_truth_stream = json_data["_understand_ai_t4_format/coordinate_system_radar_stream_raillabel"]

    assert output_stream == ground_truth_stream

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict(json_data):
    input_data = json_data["_understand_ai_t4_format/segmentation_3d"]
    input_data["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_lidar"]
    segmentation_3d = uai_format.Segmentation3d.fromdict(input_data)

    assert segmentation_3d.id == UUID(input_data["id"])
    assert segmentation_3d.object_id == UUID(input_data["objectId"])
    assert segmentation_3d.class_name == input_data["className"]
    assert segmentation_3d.associated_points == input_data["geometry"]["associatedPoints"]
    assert segmentation_3d.number_of_points == input_data["geometry"]["numberOfPointsInBox"]
    assert segmentation_3d.attributes == input_data["attributes"]
    assert segmentation_3d.sensor == uai_format.SensorReference.fromdict(input_data["sensor"])

def test_to_raillabel(json_data):
    input_data = json_data["_understand_ai_t4_format/segmentation_3d"]
    input_data["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_lidar"]
    segmentation_3d = uai_format.Segmentation3d.fromdict(input_data)
    output_data, object_id, class_name, sensor_ref = segmentation_3d.to_raillabel()
    ground_truth = json_data["_understand_ai_t4_format/segmentation_3d_raillabel"]

    assert output_data == ground_truth
    assert object_id == input_data["objectId"]
    assert class_name == "catenary_pole"
    assert sensor_ref == segmentation_3d.sensor.to_raillabel()[1]


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import typing as t
import uuid
import warnings
from dataclasses import dataclass, field

from .._util._warning import _warning
from .num import Num
from .object import Object
from .object_data import ObjectData
from .sensor import Sensor
from .sensor_reference import SensorReference


@dataclass
class Frame:
    """A container of dynamic, timewise, information.

    Parameters
    ----------
    uid: int
        Number of the frame within the annotation file. Must be unique.
    timestamp: decimal.Decimal, optional
        Timestamp containing the Unix epoch time of the frame with up to nanosecond precision.
    sensors: dict of raillabel.format.SensorReference, optional
        References to the sensors with frame specific information like timestamp and uri.
        Default is {}.
    frame_data: dict, optional
        Dictionary containing data directly connected to the frame and not to any object, like
        gps/imu data. Dictionary keys are the ID-strings of the variable the data belongs to.
        Default is {}.
    object_data: dict of raillabel.format.ObjectData, optional
        Dictionary containing the annotations per object. Dictionary keys are the object uids.
        Default is {}.

    Read-Only Attributes
    --------------------
    annotations: dict
        Consolidation of all annotations in the object_data. Only use this field if you don't need
        object tracking information. Dictionary keys are annotation UIDs.
    """

    uid: int
    timestamp: t.Optional[decimal.Decimal] = None
    sensors: t.Dict[str, SensorReference] = field(default_factory=dict)
    frame_data: t.Dict[str, Num] = field(default_factory=dict)
    object_data: t.Dict[uuid.UUID, ObjectData] = field(default_factory=dict)

    @property
    def annotations(self) -> t.Dict[uuid.UUID, t.Any]:
        """Return dict containing all annotations of this frame.

        Only use this field if you don't need object tracking
        information. Dictionary keys are annotation UIDs.
        """
        annotations = {}
        for object in self.object_data.values():
            annotations.update(object.annotations)

        return annotations

    @property
    def data(self) -> t.Dict[str, Num]:
        """Return frame data under alias.

        Will be deprecated soon.
        """
        warnings.warn(
            "data is a deprecated field and will be removed soon. Use Frame.frame_data instead.",
            category=DeprecationWarning,
        )
        return self.frame_data

    @classmethod
    def fromdict(
        cls,
        uid: str,
        data_dict: dict,
        objects: t.Dict[str, Object],
        sensors: t.Dict[str, Sensor],
        annotation_classes: dict,
    ) -> "Frame":
        """Generate a Frame object from a dict.

        Parameters
        ----------
        uid: str
            Unique identifier of the frame.
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        objects: dict
            Dictionary of all objects in the scene.
        sensors: dict
            Dictionary of all sensors in the scene.
        annotation_classes: dict
            Dictionary conaining all of the annotation classes as values with the OpenLABEL identifiers as keys.

        Returns
        -------
        frame: raillabel.format.Frame
            Converted Frame object.
        """

        frame = Frame(
            uid=int(uid),
            timestamp=cls._timestamp_fromdict(data_dict),
            sensors=cls._sensors_fromdict(data_dict, int(uid), sensors),
            frame_data=cls._frame_data_fromdict(data_dict, int(uid), annotation_classes, sensors),
            object_data=cls._objects_fromdict(
                data_dict, int(uid), objects, sensors, annotation_classes
            ),
        )

        frame = cls._fix_sensor_uri_attribute(frame)

        return frame

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {}

        if self.timestamp is not None or self.sensors != {}:
            dict_repr["frame_properties"] = {}

        if self.timestamp is not None:
            dict_repr["frame_properties"]["timestamp"] = str(self.timestamp)

        if self.sensors != {}:
            dict_repr["frame_properties"]["streams"] = {
                str(k): v.asdict() for k, v in self.sensors.items()
            }

        if self.frame_data != {}:
            dict_repr["frame_properties"]["frame_data"] = {
                "num": [v.asdict() for v in self.frame_data.values()]
            }

        if self.object_data != {}:
            dict_repr["objects"] = {str(k): v.asdict() for k, v in self.object_data.items()}

        return dict_repr

    def _timestamp_fromdict(data_dict: dict) -> t.Optional[decimal.Decimal]:

        if "frame_properties" not in data_dict or "timestamp" not in data_dict["frame_properties"]:
            return None

        return decimal.Decimal(data_dict["frame_properties"]["timestamp"])

    def _sensors_fromdict(
        data_dict: dict, frame_uid: int, scene_sensors: t.Dict[str, Sensor]
    ) -> t.Dict[str, SensorReference]:

        if "frame_properties" not in data_dict or "streams" not in data_dict["frame_properties"]:
            return {}

        sensors = {}

        for sensor_id, sensor_dict in data_dict["frame_properties"]["streams"].items():
            if sensor_id not in scene_sensors:
                _warning(
                    f"{sensor_id} does not exist as a stream, but is referenced in the "
                    + f"sync of frame {frame_uid}."
                )
                continue

            sensors[sensor_id] = SensorReference.fromdict(
                data_dict=sensor_dict, sensor=scene_sensors[sensor_id]
            )

        return sensors

    def _frame_data_fromdict(
        data_dict: dict, frame_id: int, annotation_classes: dict, sensors: t.Dict[str, Sensor]
    ) -> t.Dict[str, Num]:

        if "frame_properties" not in data_dict or "frame_data" not in data_dict["frame_properties"]:
            return {}

        frame_data = {}

        for ann_type in data_dict["frame_properties"]["frame_data"]:

            if ann_type not in annotation_classes:
                _warning(
                    f"Annotation type {ann_type} (frame {frame_id}, frame data) is "
                    + "currently not supported. Supported annotation types: "
                    + str(list(annotation_classes.keys()))
                )
                continue

            for ann_raw in data_dict["frame_properties"]["frame_data"][ann_type]:

                if "uid" not in ann_raw:
                    ann_raw["uid"] = uuid.uuid4()

                frame_data[ann_raw["name"]] = annotation_classes[ann_type].fromdict(
                    ann_raw, sensors
                )

        return frame_data

    def _objects_fromdict(
        data_dict: dict,
        frame_id: int,
        objects: t.Dict[str, Object],
        sensors: t.Dict[str, Sensor],
        annotation_classes: dict,
    ) -> t.Dict[uuid.UUID, ObjectData]:

        if "objects" not in data_dict:
            return {}

        object_data = {}

        for obj_id, obj_ann in data_dict["objects"].items():

            if obj_id not in objects:
                _warning(
                    f"{obj_id} does not exist as an object, but is referenced in the object"
                    + f" annotation of frame {frame_id}."
                )
                continue

            object_data[obj_id] = ObjectData.fromdict(
                uid=obj_id,
                data_dict=obj_ann["object_data"],
                objects=objects,
                sensors=sensors,
                annotation_classes=annotation_classes,
            )

        return object_data

    def _fix_sensor_uri_attribute(frame: "Frame") -> "Frame":

        for ann_id, ann in list(frame.annotations.items()):
            for attr_name, attr_val in ann.attributes.items():

                if attr_name != "uri":
                    continue

                _warning(
                    f"Deprecated attribute 'uri' detected in annotation {ann_id}. The error has"
                    + " been fixed. Please update the file via 'raillabel.save()'."
                )

                frame.sensors[ann.sensor.uid].uri = attr_val
                del frame.annotations[ann_id].attributes[attr_name]
                break

        return frame

    def __eq__(self, other) -> bool:
        """Handel equal comparisons."""

        if not hasattr(other, "__dict__"):
            return False

        if len(self.__dict__) != len(other.__dict__):
            return False

        for attr in self.__dict__:

            if type(getattr(self, attr)) == type(self):
                if getattr(self, attr).uid != getattr(other, attr).uid:
                    return False

            else:
                if getattr(self, attr) != getattr(other, attr):
                    return False

        return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass


@dataclass
class Quaternion:
    """A quaternion.

    Parameters
    ----------
    x: float or int
        The x component of the quaternion.
    y: float or int
        The y component of the quaternion.
    z: float or int
        The z component of the quaternion.
    w: float or int
        The w component of the quaternion.
    """

    x: t.Union[float, int]
    y: t.Union[float, int]
    z: t.Union[float, int]
    w: t.Union[float, int]



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t

from ._filter_abc import _Annotation, _FilterABC


class _FilterSensors(_FilterABC):

    PARAMETERS = ["include_sensors", "exclude_sensors"]
    LEVELS = ["frame_data", "annotation"]

    def passes_filter(self, annotation: t.Type[_Annotation]) -> bool:

        if self.include_sensors is not None:
            return annotation.sensor.uid in self.include_sensors

        elif self.exclude_sensors is not None:
            return annotation.sensor.uid not in self.exclude_sensors

        else:
            return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import typing as t
from importlib import import_module
from inspect import isclass
from io import StringIO
from pathlib import Path
from pkgutil import iter_modules

from .. import exceptions, format
from ..format._annotation import _Annotation
from ._loader_abc import LoaderABC


class LoaderRailLabelV2(LoaderABC):
    """Loader class for the OpenLabel v1 annotation format.

    Attributes
    ----------
    scene: raillabel.Scene
        Loaded raillabel.Scene with the data.
    warnings: t.List[str]
        List of warning strings, that have been found during the execution of load().
    """

    scene: format.Scene
    warnings: t.List[str]

    SCHEMA_PATH: Path = Path(__file__).parent.parent / "schemas" / "raillabel_v2_schema.json"

    @property
    def subschema_version(self):
        """Return subschema version."""
        with self.SCHEMA_PATH.open() as schema_file:
            subschema_version = json.load(schema_file)["version"]
        return subschema_version

    def load(self, data: dict, validate: bool = True) -> format.Scene:
        """Load the data into a raillabel.Scene and return it.

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.
        validate: bool
            If True, the annotation data is validated via the respective schema. This is highly
            recommended, as not validating the data may lead to Errors during loading or while handling
            the scene. However, validating may increase the loading time. Default is True.

        Returns
        -------
        scene: raillabel.Scene
            The loaded scene with the data.

        Raises
        ------
        raillabel.exceptions.SchemaError
            if validate is True and the data does not validate against the schema.
        """

        self._set_up_logger()

        if validate:
            self.validate(data)

        data = self._prepare_data(data)

        self.scene = format.Scene(
            metadata=format.Metadata.fromdict(
                data_dict=data["metadata"], subschema_version=self.subschema_version
            )
        )

        self._check_sensor_completeness(data["coordinate_systems"], data["streams"])

        for stream_id in data["streams"]:
            self.scene.sensors[stream_id] = format.Sensor.fromdict(
                uid=stream_id,
                cs_data_dict=data["coordinate_systems"][stream_id],
                stream_data_dict=data["streams"][stream_id],
            )

        for object_id in data["objects"]:
            self.scene.objects[object_id] = format.Object.fromdict(
                data["objects"][object_id], object_id
            )

        annotation_classes = self._fetch_annotation_classes()
        for frame_id in data["frames"]:
            self.scene.frames[int(frame_id)] = format.Frame.fromdict(
                uid=frame_id,
                data_dict=data["frames"][frame_id],
                objects=self.scene.objects,
                sensors=self.scene.sensors,
                annotation_classes=annotation_classes,
            )

        self.warnings = self._get_warnings()
        self._clear_log_handler()

        return self.scene

    def supports(self, data: dict) -> bool:
        """Determine if the loader is suitable for the data (lightweight).

        Parameters
        ----------
        data: dict
            A dictionary loaded from a JSON-file.

        Returns
        -------
        bool:
            If True, the Loader class is suitable for the data.
        """

        if "openlabel" not in data or "metadata" not in data["openlabel"]:
            return False

        if "subschema_version" in data["openlabel"]["metadata"]:
            return (
                "openlabel" in data
                and "metadata" in data["openlabel"]
                and "schema_version" in data["openlabel"]["metadata"]
                and data["openlabel"]["metadata"]["subschema_version"].startswith("2.")
            )

        else:
            return (
                "openlabel" in data
                and "metadata" in data["openlabel"]
                and "schema_version" in data["openlabel"]["metadata"]
            )

    def _set_up_logger(self) -> t.Tuple[StringIO, logging.StreamHandler]:
        """Set up the warnings logger.

        Returns
        -------
        StringIO
            stream containing the warnings.
        """

        logger = logging.getLogger("loader_warnings")
        warnings_stream = StringIO()
        handler = logging.StreamHandler(warnings_stream)
        handler.setLevel(logging.WARNING)
        logger.addHandler(handler)

        return warnings_stream

    def _prepare_data(self, data: dict) -> dict:
        """Add optional fields to dict to simplify interaction.

        Parameters
        ----------
        data : dict
            JSON data.

        Returns
        -------
        dict
            Enhanced JSON data.
        """

        if "coordinate_systems" not in data["openlabel"]:
            data["openlabel"]["coordinate_systems"] = {}

        if "streams" not in data["openlabel"]:
            data["openlabel"]["streams"] = {}

        if "objects" not in data["openlabel"]:
            data["openlabel"]["objects"] = {}

        if "frames" not in data["openlabel"]:
            data["openlabel"]["frames"] = {}

        return data["openlabel"]

    def _check_sensor_completeness(self, cs_data: dict, stream_data: dict):
        """Check for corresponding cs and stream completeness.

        Parameters
        ----------
        cs_data : dict
            Coordinate system data in the RailLabel format.
        stream_data : dict
            Stream data in the RailLabel format.

        Raises
        ------
        raillabel.exceptions.MissingCoordinateSystemError
            if a stream has no corresponding coordinate system.
        raillabel.exceptions.MissingStreamError
            if a coordinate system has no corresponding stream.
        raillabel.exceptions.UnsupportedParentError
            if a coordinate system has no corresponding stream.
        """

        for stream_uid in stream_data:
            if stream_uid not in cs_data:
                raise exceptions.MissingCoordinateSystemError(
                    f"Stream {stream_uid} has no corresponding coordinate system."
                )

        for cs_uid in cs_data:
            if cs_uid == "base":
                continue

            if cs_data[cs_uid]["parent"] != "base":
                raise exceptions.UnsupportedParentError(
                    f"Only 'base' is permitted as a parent for coordinate system {cs_uid}, "
                    + f"not {cs_data[cs_uid]['parent']}."
                )

            if cs_uid not in stream_data:
                raise exceptions.MissingStreamError(
                    f"Coordinate sytem {cs_uid} has no corresponding stream."
                )

    def _fetch_annotation_classes(self) -> dict:

        annotation_classes = {}

        package_dir = str(Path(__file__).resolve().parent.parent / "format")
        for (_, module_name, _) in iter_modules([package_dir]):

            module = import_module(f"raillabel.format.{module_name}")
            for attribute_name in dir(module):
                attribute = getattr(module, attribute_name)

                if (
                    isclass(attribute)
                    and issubclass(attribute, _Annotation)
                    and attribute != _Annotation
                ):
                    annotation_classes[attribute.OPENLABEL_ID] = attribute

        return annotation_classes

    def _get_warnings(self) -> t.List[str]:
        """Fetch warnings from logger as list.

        Returns
        -------
        list of str
            List of warnings.
        """

        logger = logging.getLogger("loader_warnings")
        stream = logger.handlers[-1].stream
        stream.seek(0)

        warnings_list = stream.getvalue().split("\n")

        if len(warnings_list) > 0:
            warnings_list = warnings_list[:-1]

        return warnings_list

    def _clear_log_handler(self):
        logging.getLogger("loader_warnings").handlers = []



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict(json_data):
    input_data = json_data["_understand_ai_t4_format/bounding_box_2d"]
    input_data["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    bounding_box = uai_format.BoundingBox2d.fromdict(input_data)

    assert bounding_box.id == UUID(input_data["id"])
    assert bounding_box.object_id == UUID(input_data["objectId"])
    assert bounding_box.class_name == input_data["className"]
    assert bounding_box.x_min == input_data["geometry"]["xMin"]
    assert bounding_box.y_min == input_data["geometry"]["yMin"]
    assert bounding_box.x_max == input_data["geometry"]["xMax"]
    assert bounding_box.y_max == input_data["geometry"]["yMax"]
    assert bounding_box.attributes == input_data["attributes"]
    assert bounding_box.sensor == uai_format.SensorReference.fromdict(input_data["sensor"])

def test_to_raillabel(json_data):
    input_data = json_data["_understand_ai_t4_format/bounding_box_2d"]
    input_data["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    bounding_box = uai_format.BoundingBox2d.fromdict(input_data)
    output_data, object_id, class_name, sensor_ref = bounding_box.to_raillabel()
    ground_truth = json_data["_understand_ai_t4_format/bounding_box_2d_raillabel"]

    assert output_data["name"] == ground_truth["name"]
    assert output_data["val"] == ground_truth["val"]
    assert output_data["attributes"] == ground_truth["attributes"]
    assert object_id == input_data["objectId"]
    assert class_name == input_data["className"]
    assert sensor_ref == bounding_box.sensor.to_raillabel()[1]

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from importlib import metadata as importlib_metadata


@dataclass
class Metadata:
    """Container for metadata information about the scene itself.

    Parameters
    ----------
    schema_version: str
        Version number of the OpenLABEL schema this annotation object follows.
    annotator: str, optional
        Name or description of the annotator that created the annotations. Default is None.
    comment: str, optional
        Additional information or description about the annotation content. Default is None.
    exporter_version: str, optional
        Version of the raillabel-devkit, that last exported the scene. Default is None.
    file_version: str, optional
        Version number of the raillabel annotation content. Default is None.
    name: str, optional
        Name of the raillabel annotation content. Default is None.
    subschema_version: str, optional
        Version number of the RailLabel schema this annotation object follows. Default is None.
    tagged_file: str, optional
        Directory with the exported data_dict (e.g. images, point clouds). Default is None.
    """

    schema_version: str
    annotator: t.Optional[str] = None
    comment: t.Optional[str] = None
    exporter_version: t.Optional[str] = None
    file_version: t.Optional[str] = None
    name: t.Optional[str] = None
    subschema_version: t.Optional[str] = None
    tagged_file: t.Optional[str] = None

    @classmethod
    def fromdict(cls, data_dict: dict, subschema_version: t.Optional[str] = None) -> "Metadata":
        """Generate a Metadata object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        subschema_version: str, optional
            Version of the RailLabel subschema

        Returns
        -------
        metadata: Metadata
            Converted metadata.
        warnings: list of str
            List of non-critical errors, that have occurred during the conversion.
        """

        return Metadata(
            schema_version=data_dict["schema_version"],
            subschema_version=subschema_version,
            annotator=data_dict.get("annotator"),
            file_version=data_dict.get("file_version"),
            name=data_dict.get("name"),
            tagged_file=data_dict.get("tagged_file"),
            comment=data_dict.get("comment"),
            exporter_version=cls._collect_exporter_version(),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {"schema_version": str(self.schema_version)}

        if self.annotator is not None:
            dict_repr["annotator"] = str(self.annotator)

        if self.comment is not None:
            dict_repr["comment"] = str(self.comment)

        if self.exporter_version is not None:
            dict_repr["exporter_version"] = str(self.exporter_version)

        if self.file_version is not None:
            dict_repr["file_version"] = str(self.file_version)

        if self.name is not None:
            dict_repr["name"] = str(self.name)

        if self.subschema_version is not None:
            dict_repr["subschema_version"] = str(self.subschema_version)

        if self.tagged_file is not None:
            dict_repr["tagged_file"] = str(self.tagged_file)

        return dict_repr

    @classmethod
    def _collect_exporter_version(cls) -> t.Optional[str]:

        try:
            exporter_version = importlib_metadata.version("raillabel")
        except importlib_metadata.PackageNotFoundError:
            return None

        version_number_length = len(exporter_version) - len(exporter_version.split(".")[-1])
        return exporter_version[: version_number_length - 1]



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Quaternion:
    """Dimensional information of an object in 3d.

    Parameters
    ----------
    x: float
        The x component of the quaternion.
    y: float
        The y component of the quaternion.
    z: float
        The z component of the quaternion.
    w: float
        The w component of the quaternion.
    """

    x: float
    y: float
    z: float
    w: float

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Quaternion":
        """Generate a Quaternion from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Quaternion
            Converted quaternion.
        """

        return Quaternion(
            x=float(data_dict["x"]),
            y=float(data_dict["y"]),
            z=float(data_dict["z"]),
            w=float(data_dict["w"]),
        )



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t

from ._filter_abc import _Annotation, _FilterABC


class _FilterAnnotationIds(_FilterABC):

    PARAMETERS = ["include_annotation_ids", "exclude_annotation_ids"]
    LEVELS = ["annotation"]

    def passes_filter(self, annotation: t.Type[_Annotation]) -> bool:

        if self.include_annotation_ids is not None:
            return annotation.uid in self.include_annotation_ids

        elif self.exclude_annotation_ids is not None:
            return annotation.uid not in self.exclude_annotation_ids

        else:
            return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t

from ._filter_abc import _Annotation, _FilterABC


class _FilterAttributes(_FilterABC):

    PARAMETERS = ["include_attributes", "exclude_attributes"]
    LEVELS = ["annotation"]

    def passes_filter(self, annotation: t.Type[_Annotation]) -> bool:

        if self.include_attributes is not None:
            for attribute_id, attribute_val in self.include_attributes.items():

                if attribute_val is None:
                    if attribute_id not in annotation.attributes:
                        return False

                else:
                    if (
                        attribute_id not in annotation.attributes
                        or annotation.attributes[attribute_id] != attribute_val
                    ):
                        return False

        elif self.exclude_attributes is not None:

            for attribute_id, attribute_val in self.exclude_attributes.items():

                if attribute_val is None:
                    if attribute_id in annotation.attributes:
                        return False

                else:
                    if (
                        attribute_id in annotation.attributes
                        and attribute_val == annotation.attributes[attribute_id]
                    ):
                        return False

        return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Package containing the loader classes for all supported formats."""

from importlib import import_module
from inspect import isclass
from pathlib import Path
from pkgutil import iter_modules

# iterate through the modules in the current package
package_dir = str(Path(__file__).resolve().parent)
for (_, module_name, _) in iter_modules([package_dir]):

    # import the module and iterate through its attributes
    module = import_module(f"{__name__}.{module_name}")
    for attribute_name in dir(module):
        attribute = getattr(module, attribute_name)

        if isclass(attribute):
            # Add the class to this package's variables
            globals()[attribute_name] = attribute



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict():
    input_data = {
        "height": 12.887274595406309,
        "width": 0.9726718154765793,
        "length": 0.7487449536720978

    }
    size = uai_format.Size3d.fromdict(input_data)

    assert size.height == float(input_data["height"])
    assert size.width == float(input_data["width"])
    assert size.length == float(input_data["length"])

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from decimal import Decimal
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def _prepare_frame_data(frame_data: dict, json_data: dict) -> dict:
    frame_data["annotations"]["2D_BOUNDING_BOX"] = [json_data["_understand_ai_t4_format/bounding_box_2d"]]
    frame_data["annotations"]["2D_BOUNDING_BOX"][0]["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    frame_data["annotations"]["2D_POLYLINE"] = [json_data["_understand_ai_t4_format/polyline_2d"]]
    frame_data["annotations"]["2D_POLYLINE"][0]["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    frame_data["annotations"]["2D_POLYGON"] = [json_data["_understand_ai_t4_format/polygon_2d"]]
    frame_data["annotations"]["2D_POLYGON"][0]["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    frame_data["annotations"]["3D_BOUNDING_BOX"] = [json_data["_understand_ai_t4_format/bounding_box_3d"]]
    frame_data["annotations"]["3D_BOUNDING_BOX"][0]["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_lidar"]
    frame_data["annotations"]["3D_SEGMENTATION"] = [json_data["_understand_ai_t4_format/segmentation_3d"]]
    frame_data["annotations"]["3D_SEGMENTATION"][0]["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_lidar"]

    return frame_data


def _prepare_ground_truth(ground_truth: dict, json_data:dict) -> dict:
    ground_truth["objects"]["48c988bd-76f1-423f-b46d-7e7acb859f31"]["object_data"]["bbox"] = [
        json_data["_understand_ai_t4_format/bounding_box_2d_raillabel"]
    ]
    ground_truth["objects"]["48c988bd-76f1-423f-b46d-7e7acb859f31"]["object_data"]["cuboid"] = [
        json_data["_understand_ai_t4_format/bounding_box_3d_raillabel"]
    ]
    ground_truth["objects"]["58e7edd8-a7ee-4775-a837-e6dd375e8150"]["object_data"]["poly2d"] = [
        json_data["_understand_ai_t4_format/polygon_2d_raillabel"]
    ]
    ground_truth["objects"]["4d8eca35-6c1d-4159-8062-21c2f2c051df"]["object_data"]["poly2d"] = [
        json_data["_understand_ai_t4_format/polyline_2d_raillabel"]
    ]
    ground_truth["objects"]["05a7e7a7-91e1-49ef-a172-780f2461f013"]["object_data"]["vec"] = [
        json_data["_understand_ai_t4_format/segmentation_3d_raillabel"]
    ]

    return ground_truth


def test_fromdict(json_data):
    input_data = _prepare_frame_data(json_data["_understand_ai_t4_format/frame"], json_data)
    frame = uai_format.Frame.fromdict(input_data)

    assert frame.id == int(input_data["frameId"])
    assert frame.timestamp == Decimal(input_data["timestamp"])
    assert frame.bounding_box_2ds[json_data["_understand_ai_t4_format/bounding_box_2d"]["id"]] == uai_format.BoundingBox2d.fromdict(
        json_data["_understand_ai_t4_format/bounding_box_2d"]
    )
    assert frame.polyline_2ds[json_data["_understand_ai_t4_format/polyline_2d"]["id"]] == uai_format.Polyline2d.fromdict(
        json_data["_understand_ai_t4_format/polyline_2d"]
    )
    assert frame.polygon_2ds[json_data["_understand_ai_t4_format/polygon_2d"]["id"]] == uai_format.Polygon2d.fromdict(
        json_data["_understand_ai_t4_format/polygon_2d"]
    )
    assert frame.bounding_box_3ds[json_data["_understand_ai_t4_format/bounding_box_3d"]["id"]] == uai_format.BoundingBox3d.fromdict(
        json_data["_understand_ai_t4_format/bounding_box_3d"]
    )
    assert frame.segmentation_3ds[json_data["_understand_ai_t4_format/segmentation_3d"]["id"]] == uai_format.Segmentation3d.fromdict(
        json_data["_understand_ai_t4_format/segmentation_3d"]
    )

def test_to_raillabel(json_data):
    input_data = _prepare_frame_data(json_data["_understand_ai_t4_format/frame"], json_data)
    frame = uai_format.Frame.fromdict(input_data)
    output_data = frame.to_raillabel()
    ground_truth = _prepare_ground_truth(json_data["_understand_ai_t4_format/frame_raillabel"], json_data)

    assert output_data == ground_truth

def test_translated_sensors(json_data):
    input_data = _prepare_frame_data(json_data["_understand_ai_t4_format/frame"], json_data)
    frame = uai_format.Frame.fromdict(input_data)
    ground_truth_streams = json_data["_understand_ai_t4_format/frame_raillabel"]["frame_properties"]["streams"]

    assert frame.translated_sensors.keys() == ground_truth_streams.keys()

def test_translated_objects(json_data):
    input_data = _prepare_frame_data(json_data["_understand_ai_t4_format/frame"], json_data)
    frame = uai_format.Frame.fromdict(input_data)
    ground_truth_objects = json_data["_understand_ai_t4_format/frame_raillabel"]["objects"]

    assert frame.translated_objects.keys() == ground_truth_objects.keys()


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class FrameInterval:
    """Closed interval of frames.

    Parameters
    ----------
    frame_start: int
        Initial frame number of the interval (inclusive).
    frame_end: int
        Ending frame number of the interval (inclusive).
    """

    frame_start: int
    frame_end: int

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {
            "frame_start": int(self.frame_start),
            "frame_end": int(self.frame_end),
        }

        return dict_repr

    def __len__(self) -> int:
        """Return the length in frames."""
        return abs(self.frame_start - self.frame_end) + 1



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict(json_data):
    input_data = json_data["_understand_ai_t4_format/polygon_2d"]
    input_data["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    polygon = uai_format.Polygon2d.fromdict(input_data)

    assert polygon.id == UUID(input_data["id"])
    assert polygon.object_id == UUID(input_data["objectId"])
    assert polygon.class_name == input_data["className"]
    assert polygon.points == [(p[0], p[1]) for p in input_data["geometry"]["points"]]
    assert polygon.attributes == input_data["attributes"]
    assert polygon.sensor == uai_format.SensorReference.fromdict(input_data["sensor"])

def test_to_raillabel(json_data):
    input_data = json_data["_understand_ai_t4_format/polygon_2d"]
    input_data["sensor"] = json_data["_understand_ai_t4_format/sensor_reference_camera"]
    polygon = uai_format.Polygon2d.fromdict(input_data)
    output_data, object_id, class_name, sensor_ref = polygon.to_raillabel()
    ground_truth = json_data["_understand_ai_t4_format/polygon_2d_raillabel"]

    assert output_data["name"] == ground_truth["name"]
    assert output_data["val"] == ground_truth["val"]
    assert output_data["attributes"] == ground_truth["attributes"]
    assert output_data["closed"] == True
    assert output_data["mode"] == "MODE_POLY2D_ABSOLUTE"
    assert object_id == input_data["objectId"]
    assert class_name == "signal_pole"
    assert sensor_ref == polygon.sensor.to_raillabel()[1]


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ._annotation import _Annotation
from .point2d import Point2d


@dataclass
class Poly2d(_Annotation):
    """Sequence of 2D points. Can either be a polygon or polyline.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier for the annotation.
    name: str
        Human readable name describing the annotation.
    points: list of raillabel.format.Point2d
        List of the 2d points that make up the polyline.
    closed: bool
        This parameter states, whether the polyline represents a closed shape (a polygon) or an
        open line.
    mode: str, optional
        Mode of the polyline list of values: "MODE_POLY2D_ABSOLUTE" determines that the poly2d list
        contains the sequence of (x, y) values of all points of the polyline. "MODE_POLY2D_RELATIVE"
        specifies that only the first point of the sequence is defined with its (x, y) values, while
        all the rest are defined relative to it. "MODE_POLY2D_SRF6DCC" specifies that SRF6DCC chain
        code method is used. "MODE_POLY2D_RS6FCC" specifies that the RS6FCC method is used. Default
        is 'MODE_POLY2D_ABSOLUTE'.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the name str of the attribute, values are the
        attribute values. Default is {}.
    sensor: raillabel.format.CoordinateSystem, optional
        A reference to the sensor, this annotation is labeled in. Default is None.
    """

    points: t.List[Point2d] = None
    closed: bool = None
    mode: str = "MODE_POLY2D_ABSOLUTE"

    OPENLABEL_ID = "poly2d"
    _REQ_FIELDS = ["points", "closed"]

    @classmethod
    def fromdict(
        self,
        data_dict: dict,
        sensors: dict,
    ) -> "Poly2d":
        """Generate a Poly2d object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensors: dict
            Dictionary containing all sensors for the scene.

        Returns
        -------
        annotation: Poly2d
            Converted annotation.
        """

        return Poly2d(
            uid=str(data_dict["uid"]),
            name=str(data_dict["name"]),
            closed=data_dict["closed"],
            mode=data_dict["mode"],
            points=self._points_fromdict(data_dict),
            sensor=self._coordinate_system_fromdict(data_dict, sensors),
            attributes=self._attributes_fromdict(data_dict),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = self._annotation_required_fields_asdict()

        dict_repr["closed"] = bool(self.closed)
        dict_repr["val"] = []
        dict_repr["mode"] = self.mode
        for point in self.points:
            dict_repr["val"].extend(point.asdict())

        dict_repr.update(self._annotation_optional_fields_asdict())

        return dict_repr

    def _points_fromdict(data_dict: dict) -> t.List[Point2d]:
        points = []
        for i in range(0, len(data_dict["val"]), 2):
            points.append(Point2d(x=data_dict["val"][i], y=data_dict["val"][i + 1]))
        return points



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


def test_post_init():
    bbox = raillabel.format.Bbox(
        uid="d2764400-8560-4991-a491-ada598b345c8",
        name="test_name",
        pos=raillabel.format.Point2d(0, 1),
        size=raillabel.format.Size2d(2, 3),
    )

    with pytest.raises(TypeError):
        bbox = raillabel.format.Bbox(
            uid="d2764400-8560-4991-a491-ada598b345c8",
            name="test_name",
            pos=raillabel.format.Point2d(0, 1),
        )


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass, field

from .._util._warning import _warning
from .sensor import Sensor


@dataclass
class _Annotation(ABC):

    uid: str
    name: str
    attributes: t.Dict[str, t.Union[int, float, bool, str, list]] = field(default_factory=dict)
    sensor: Sensor = None

    @property
    @abstractproperty
    def _REQ_FIELDS(self) -> t.List[str]:
        raise NotImplementedError

    @property
    @abstractproperty
    def OPENLABEL_ID(self) -> t.List[str]:
        raise NotImplementedError

    # === Public Methods =====================================================

    @abstractmethod
    def asdict(self) -> t.Dict:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def fromdict(self, data_dict: t.Dict, sensors: t.Dict) -> t.Type["_Annotation"]:
        raise NotImplementedError

    # === Private Methods ====================================================

    def _annotation_required_fields_asdict(self) -> t.Dict:
        """Return the required fields from the parent class to dict."""
        return {
            "uid": str(self.uid),
            "name": str(self.name),
        }

    def _annotation_optional_fields_asdict(self) -> t.Dict:
        """Return the optional fields from the parent class to dict."""

        dict_repr = {}

        if self.sensor is not None:
            dict_repr["coordinate_system"] = str(self.sensor.uid)

        if self.attributes != {}:
            dict_repr["attributes"] = {}

            for attr_name, attr_value in self.attributes.items():

                # Since the annotation stores the attributes in a collective
                # dictionary, they must be seperated by type in order to comply
                # with the OpenLabel format.

                if type(attr_value) == str:
                    attr_type = "text"

                elif type(attr_value) in [float, int]:
                    attr_type = "num"

                elif type(attr_value) == bool:
                    attr_type = "boolean"

                elif type(attr_value) in [list, tuple]:
                    attr_type = "vec"

                else:
                    raise TypeError(
                        f"Attribute type {attr_value.__class__.__name__} of {attr_value} is not "
                        + "supported. Supported types are str, float, int, bool, list, tuple."
                    )

                if attr_type not in dict_repr["attributes"]:
                    dict_repr["attributes"][attr_type] = []

                dict_repr["attributes"][attr_type].append({"name": attr_name, "val": attr_value})

        return dict_repr

    def _coordinate_system_fromdict(data_dict: dict, sensors: dict) -> t.Optional[Sensor]:

        is_coordinate_system_in_data = (
            "coordinate_system" in data_dict and data_dict["coordinate_system"] != ""
        )

        if not is_coordinate_system_in_data:
            return None

        if data_dict["coordinate_system"] not in sensors:
            _warning(
                f"{data_dict['coordinate_system']} does not exist as a coordinate system, "
                + f"but is referenced for the annotation {data_dict['uid']}."
            )
            return None

        return sensors[data_dict["coordinate_system"]]

    def _attributes_fromdict(
        data_dict: dict,
    ) -> t.Dict[str, t.Union[int, float, bool, str, list]]:

        if "attributes" not in data_dict:
            return {}

        return {a["name"]: a["val"] for l in data_dict["attributes"].values() for a in l}

    # === Special Methods ====================================================

    def __post_init__(self):
        """Check for required arguments after __init__.

        Inheritance in dataclasses has the flaw, that when the parent class has fields with
        defaults and the child class has fields without, a TypeError is raised due to required
        fields following optional ones. This is solved by making all fields in the child
        optional and checking for required fields in __post_init__().

        Raises
        ------
        TypeError
            If a required field has not been set.
        """

        for f in self._REQ_FIELDS:
            if getattr(self, f) is None:
                raise TypeError(f"{f} is a required argument for {self.__class__.__name__}")



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from ._filter_abc import Frame, _FilterABC


class _FilterFrame(_FilterABC):

    PARAMETERS = ["include_frames", "exclude_frames"]
    LEVELS = ["frame"]

    def passes_filter(self, frame: Frame) -> bool:

        if self.include_frames is not None:
            return int(frame.uid) in self.include_frames

        elif self.exclude_frames is not None:
            return int(frame.uid) not in self.exclude_frames

        else:
            return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t

from ._filter_abc import _Annotation, _FilterABC


class _FilterObjectIds(_FilterABC):

    PARAMETERS = ["include_object_ids", "exclude_object_ids"]
    LEVELS = ["object"]

    def passes_filter(self, object: t.Type[_Annotation]) -> bool:

        if self.include_object_ids is not None:
            return object.uid in self.include_object_ids

        elif self.exclude_object_ids is not None:
            return object.uid not in self.exclude_object_ids

        else:
            return True



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import logging


def _warning(message: str) -> logging.Logger:
    """Create a loader warning."""
    logging.getLogger("loader_warnings").warning(message)



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict():
    input_data = {
        "x": 27.075067129430366,
        "y": -3.7850908693727328,
        "z": 6.671744016124283
    }
    point = uai_format.Point3d.fromdict(input_data)

    assert point.x == float(input_data["x"])
    assert point.y == float(input_data["y"])
    assert point.z == float(input_data["z"])

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t

from ._filter_abc import _Annotation, _FilterABC


class _FilterAnnotationTypes(_FilterABC):

    PARAMETERS = ["include_annotation_types", "exclude_annotation_types"]
    LEVELS = ["annotation"]

    def passes_filter(self, annotation: t.Type[_Annotation]) -> bool:

        if self.include_annotation_types is not None:
            return annotation.__class__.__name__.lower() in self.include_annotation_types

        elif self.exclude_annotation_types is not None:
            return annotation.__class__.__name__.lower() not in self.exclude_annotation_types

        else:
            return True

    def _process_filter_args(self, filter_args):
        return [arg.lower() for arg in filter_args]



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict(json_data):
    input_data = json_data["_understand_ai_t4_format/metadata"]
    metadata = uai_format.Metadata.fromdict(input_data)

    assert metadata.clip_id == input_data["clip_id"]
    assert metadata.external_clip_id == input_data["external_clip_id"]
    assert metadata.project_id == input_data["project_id"]
    assert metadata.export_time == input_data["export_time"]
    assert metadata.exporter_version == input_data["exporter_version"]
    assert metadata.coordinate_system_3d == input_data["coordinate_system_3d"]
    assert metadata.coordinate_system_reference == input_data["coordinate_system_reference"]
    assert metadata.folder_name == input_data["folder_name"]

def test_to_raillabel(json_data):
    input_data = json_data["_understand_ai_t4_format/metadata"]
    output_data = uai_format.Metadata.fromdict(input_data).to_raillabel()
    ground_truth = json_data["_understand_ai_t4_format/metadata_raillabel"]
    ground_truth["subschema_version"] = json_data["raillabel_v2_schema"]["version"]

    assert output_data == ground_truth

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class IntrinsicsRadar:
    """Intrinsic calibration for a radar sensor.

    Parameters
    ----------
    resolution_px_per_m: float
        Factor for calculating the 3D coordinates of a pixel in the cartesian radar images. Number of pixels in the images per meter from the sensor.
    width_px: int
        Width of the cartesian image frame in pixels.
    height_px: int
        Height of the cartesian image frame in pixels.
    """

    resolution_px_per_m: float
    width_px: int
    height_px: int

    @classmethod
    def fromdict(self, data_dict: dict) -> "IntrinsicsRadar":
        """Generate a IntrinsicsRadar object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.

        Returns
        -------
        raillabel.format.IntrinsicsRadar
            Converted IntrinsicsRadar object.
        """

        return IntrinsicsRadar(
            resolution_px_per_m=data_dict["resolution_px_per_m"],
            width_px=data_dict["width_px"],
            height_px=data_dict["height_px"],
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return {
            "resolution_px_per_m": float(self.resolution_px_per_m),
            "width_px": int(self.width_px),
            "height_px": int(self.height_px),
        }



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
import os
import typing as t
from pathlib import Path

import fastjsonschema
import jsonschema

from . import exceptions


def validate(data: dict, schema_path: str = "raillabel_v2") -> t.Tuple[bool, t.List[str]]:
    """Validate JSON data represented by a dict via a given schema.

    Parameters
    ----------
    data: dict
        JSON data to be validated.
    schema_path: str, optional
        Path to the JSON schema used for the validation. If the schema is in the /schemas
        folder, the format name can be used (i.e. schema_path can be 'raillabel_v2' or
        'raillabel_v2_schema' to load the raillabel_v2_schema.json file). Default is
        'raillabel_v2'.

    Returns
    -------
    is_data_valid: bool
        True if the data validates against the schema, False if not.
    schema_errors: t.List of str
        All SchemaError messages found in the data. If the data is valid (if no SchemaErrors are
        found), this is an empty t.List.
    """

    # Since the schema_path can either be a complete path or the short name of a schema, these two
    # options must be distinguished. It is therefore assumed, that a complete path contains at #
    # least one '/' or '\'.

    if "/" in schema_path or "\\" in schema_path:  # if schema_path is a complete path
        schema_path = Path(schema_path)

    else:  # if schema_path is a schema name in /schemas
        local_schemas = [  # t.List of json files in /schemas
            p for p in os.listdir(Path(__file__).parent / "schemas") if p.endswith(".json")
        ]

        applicable_local_schemas = []  # t.List of possible schemas
        for local_schema_path in local_schemas:
            if schema_path in local_schema_path:
                applicable_local_schemas.append(local_schema_path)

        if len(applicable_local_schemas) == 1:  # if exactly one applicable file has been found
            schema_path = Path(__file__).parent / "schemas" / applicable_local_schemas[0]

        elif len(applicable_local_schemas) == 0:  # if no applicable files have been found

            err_msg = f"The key {schema_path} does not apply to a schema. Available schema files:"
            for p in local_schemas:
                err_msg += "\n - " + p

            raise FileNotFoundError(err_msg)

        else:  # if more than one applicable files have been found

            err_msg = f"The key {schema_path} applies to multiple files in /schemas:"
            for p in applicable_local_schemas:
                err_msg += "\n - " + p

            raise exceptions.AmbiguousSchemaNameError(err_msg)

    # Loads the schema data
    try:
        with schema_path.open() as schema_file:
            schema = json.load(schema_file)

    except FileNotFoundError as e:
        raise FileNotFoundError(f"The schema file could not be found in {schema_path}") from e

    # Validates the data
    schema_errors = []

    try:
        # Use fastjsonschema since its faster (duh), and use jsonschema as a fallback if we find
        # an error to maintain jsonschemas way of reporting all errors
        fastjsonschema.validate(schema, data)
    except fastjsonschema.JsonSchemaException as _:
        validator = jsonschema.Draft7Validator(schema=schema)

        for error in validator.iter_errors(data):
            schema_errors.append("$" + error.json_path[1:] + ": " + str(error.message))

    is_data_valid = len(schema_errors) == 0

    return is_data_valid, schema_errors



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json

from . import exceptions, format, format_loaders
from .format_loaders._loader_abc import LoaderABC


def load(path: str, validate: bool = False, show_warnings: bool = True) -> format.Scene:
    """Load an annotation file of any supported type.

    Parameters
    ----------
    path: str
        Path to the annotation file.
    validate: bool, optional
        If True, the annotation data is validated via the respective schema. This is highly
        recommended, as not validating the data may lead to errors during loading or while handling
        the scene. However, validating may increase the loading time. Default is False.
    show_warnings: bool, optional
        If True, any non-critical inconsistencies in the data are output as a warning. Default is
        True.

    Returns
    -------
    scene: raillabel.Scene
        Scene with the loaded data.

    Raises
    ------
    raillabel.exceptions.UnsupportedFormatError
        if the annotation file does not match any loaders.
    raillabel.exceptions.SchemaError
        if during the validation, errors in the annotation file are found.
    """

    # To expand the supported formats in a simple manner, load() automatically fetches all classes
    # in the format_loaders directory and checks if they are suitable as loaders. To avoid errors
    # caused by potential other files and classes in the directory, only classes, which inherite
    # from LoaderABC are considered.

    loader_classes = []
    for cls in format_loaders.__dict__.values():
        if isinstance(cls, type) and issubclass(cls, LoaderABC) and cls != LoaderABC:
            loader_classes.append(cls)

    # Checks for the supported file type
    if not str(path).lower().endswith(".json"):
        raise exceptions.UnsupportedFormatError(f"{path} is not in a supported file format.")

    # Loads the JSON data
    with open(path) as data_file:
        data = json.load(data_file)

    # Iterates over the loader classes to find a suitable one
    for loader_class in loader_classes:

        loader = loader_class()

        # Checks if the loader supports the data
        if not loader.supports(data):
            continue

        # Loads the scene
        scene = loader.load(data, validate=validate)

        # Outputs the warnings
        if show_warnings:

            if len(loader.warnings) > 0:
                print(f"During the loading of {path} warnings have occurred:")

                for warning in loader.warnings:
                    print(" - " + warning)

        return scene

    # This part of the code is only reached if no suitable loader has been found or the data is
    # not in a supported file format. Therefore an exception is raised.

    raise exceptions.UnsupportedFormatError(f"{path} is not in a supported file format.")



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Object:
    """Physical, unique object in the data, that can be tracked via its UID.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier for the object.
    name: str
        Name of the object. It is a friendly name and not used for indexing. Commonly the class
        name is used followed by an underscore and an integer (i.e. person_0032).
    type: str
        The type of an object defines the class the object corresponds to.
    """

    uid: str
    name: str
    type: str

    @classmethod
    def fromdict(cls, data_dict: dict, object_uid: str) -> "Object":
        """Generate a Object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        object_uid: str
            Unique identifier of the object.

        Returns
        -------
        object: raillabel.format.Object
            Converted object.
        """
        return Object(uid=object_uid, type=data_dict["type"], name=data_dict["name"])

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        return {"name": str(self.name), "type": str(self.type)}



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from ._annotation import _Annotation


@dataclass
class Num(_Annotation):
    """A number.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier of the annotation.
    name: str
        Human readable name describing the annotation.
    val: int or float
        This is the value of the number object.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the uid str of the attribute, values are the
        attribute values. Default is {}.
    sensor: raillabel.format.CoordinateSystem, optional
        A reference to the sensor, this value is represented in. Default is None.
    """

    val: t.Union[int, float] = None

    OPENLABEL_ID = "num"
    _REQ_FIELDS = ["val"]

    @classmethod
    def fromdict(self, data_dict: dict, sensors: dict) -> "Num":
        """Generate a Num object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensors: dict
            Dictionary containing all sensors for the scene.

        Returns
        -------
        annotation: Num
            Converted annotation.
        """

        return Num(
            uid=str(data_dict["uid"]),
            name=str(data_dict["name"]),
            val=data_dict["val"],
            sensor=self._coordinate_system_fromdict(data_dict, sensors),
            attributes=self._attributes_fromdict(data_dict),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = self._annotation_required_fields_asdict()

        dict_repr["val"] = self.val

        dict_repr.update(self._annotation_optional_fields_asdict())

        return dict_repr



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
from pathlib import Path


def translate_sensor_id(original_sensor_id: str) -> str:
    """Translate deprecated sensor ids to the correct ones.

    Parameters
    ----------
    original_sensor_id : str
        Original id of the sensor.

    Returns
    -------
    str
        Translated id or original_sensor_id, if no translation could be found.
    """
    return TRANSLATION["streams"].get(original_sensor_id, original_sensor_id)


def translate_class_id(original_class_id: str) -> str:
    """Translate deprecated class ids to the correct ones.

    Parameters
    ----------
    original_class_id : str
        Original id of the class.

    Returns
    -------
    str
        Translated id or original_class_id, if no translation could be found.
    """
    return TRANSLATION["classes"].get(original_class_id, original_class_id)


def fetch_sensor_type(sensor_id: str) -> str:
    """Fetch sensor type from translation file.

    Parameters
    ----------
    sensor_id : str
        Id of the sensor.

    Returns
    -------
    str
        Sensor type or 'other' if sensor_id not found in translation.json.
    """
    return TRANSLATION["stream_types"].get(sensor_id, "other")


def fetch_sensor_resolutions(sensor_id: str) -> dict:
    """Fetch sensor resolution from translation file.

    Parameters
    ----------
    sensor_id : str
        Id of the sensor.

    Returns
    -------
    dict
        Dictionary containing the resolution information. Key 'x' contains the width in pixels,
        key 'y' contains the height in pixels. If the sensor is a radar, 'resolution_px_per_m' is
        also included.
    """
    return TRANSLATION["stream_resolutions"].get(
        sensor_id, {"x": None, "y": None, "resolution_px_per_m": None}
    )


def _load_translation():
    """Load the translation file when the module is imported.

    This prevents it from beeing loaded for every annotation.
    """

    global TRANSLATION

    translatiion_path = Path(__file__).parent.parent / "format_loaders" / "translation.json"
    with translatiion_path.open() as translation_file:
        TRANSLATION = json.load(translation_file)


TRANSLATION = {}

_load_translation()



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel._understand_ai_t4_format as uai_format


def test_fromdict():
    input_data = {
        "x": -7.710484478135359e-17,
        "y": -7.978951661384013e-17,
        "z": 0.7191010536652217,
        "w": 0.6949055148849863
    }
    quaternion = uai_format.Quaternion.fromdict(input_data)

    assert quaternion.x == float(input_data["x"])
    assert quaternion.y == float(input_data["y"])
    assert quaternion.z == float(input_data["z"])
    assert quaternion.w == float(input_data["w"])

# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass


@dataclass
class Size2d:
    """The size of a rectangle in a 2d image.

    Parameters
    ----------
    x: float or int
        The size along the x-axis.
    y: float or int
        The size along the y-axis.
    """

    x: float
    y: float



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


@pytest.fixture
def loader():
    return raillabel.format_loaders.LoaderUnderstandAiT4()


def test_supports_true(json_data, loader):
    assert loader.supports(json_data["understand_ai_real_life"])


def test_supports_false(json_data, loader):
    data = json_data["understand_ai_real_life"]
    del data["metadata"]["project_id"]
    assert not loader.supports(data)


def test_load(json_data, loader):
    input_data_raillabel = remove_non_parsed_fields(json_data["openlabel_v1_short"])
    input_data_uai = json_data["understand_ai_t4_short"]

    scene_ground_truth = raillabel.format_loaders.LoaderRailLabelV2().load(input_data_raillabel, validate=False)
    scene = loader.load(input_data_uai, validate=False)

    scene.metadata = scene_ground_truth.metadata

    assert scene.asdict() == scene_ground_truth.asdict()

def remove_non_parsed_fields(raillabel_data: dict) -> dict:
    """Return RailLabel file with frame_data and poly3ds removed."""

    for frame in raillabel_data["openlabel"]["frames"].values():

        if "frame_data" in frame["frame_properties"]:
            del frame["frame_properties"]["frame_data"]

        for object_id, object in list(frame["objects"].items()):
            if "poly3d" not in object["object_data"]:
                continue

            del object["object_data"]["poly3d"]
            if len(object["object_data"]) == 0:
                del frame["objects"][object_id]

    return raillabel_data


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


@pytest.fixture
def loader():
    return raillabel.format_loaders.LoaderRailLabelV2()


def test_supports_true(json_data, loader):
    assert loader.supports(json_data["openlabel_v1_short"])


def test_supports_false(json_data, loader):
    data = json_data["openlabel_v1_short"]
    data["openlabel"]["metadata"]["subschema_version"] = "3.0.0"
    assert not loader.supports(data)


def test_load_metadata(json_data, loader):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    assert scene.metadata.annotator == ground_truth["metadata"]["annotator"]
    assert scene.metadata.comment == ground_truth["metadata"]["comment"]
    assert scene.metadata.name == ground_truth["metadata"]["name"]
    assert scene.metadata.schema_version == ground_truth["metadata"]["schema_version"]
    assert scene.metadata.tagged_file == ground_truth["metadata"]["tagged_file"]
    assert scene.metadata.subschema_version == json_data["raillabel_v2_schema"]["version"]


def test_load_sensors(json_data, loader):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    assert len(scene.sensors) == len(ground_truth["streams"])
    for sensor_id, sensor in scene.sensors.items():

        assert sensor_id in ground_truth["streams"]
        assert sensor.uid == sensor_id
        assert sensor.type.value == ground_truth["streams"][sensor_id]["type"]
        assert sensor.uri == ground_truth["streams"][sensor_id]["uri"]

        assert [
            sensor.extrinsics.pos.x,
            sensor.extrinsics.pos.y,
            sensor.extrinsics.pos.z,
        ] == ground_truth["coordinate_systems"][sensor_id]["pose_wrt_parent"]["translation"]
        assert [
            sensor.extrinsics.quat.x,
            sensor.extrinsics.quat.y,
            sensor.extrinsics.quat.z,
            sensor.extrinsics.quat.w,
        ] == ground_truth["coordinate_systems"][sensor_id]["pose_wrt_parent"]["quaternion"]

        if sensor.type.value != "camera":
            continue

        assert list(sensor.intrinsics.camera_matrix) == ground_truth["streams"][sensor_id]["stream_properties"]["intrinsics_pinhole"]["camera_matrix"]
        assert list(sensor.intrinsics.distortion) == ground_truth["streams"][sensor_id]["stream_properties"]["intrinsics_pinhole"]["distortion_coeffs"]
        assert sensor.intrinsics.width_px == ground_truth["streams"][sensor_id]["stream_properties"]["intrinsics_pinhole"]["width_px"]
        assert sensor.intrinsics.height_px == ground_truth["streams"][sensor_id]["stream_properties"]["intrinsics_pinhole"]["height_px"]


def test_load_objects(json_data, loader):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    assert len(scene.objects) == len(ground_truth["objects"])
    for object_id, object in scene.objects.items():

        assert object_id in ground_truth["objects"]
        assert object.uid == object_id
        assert object.name == ground_truth["objects"][object_id]["name"]
        assert object.type == ground_truth["objects"][object_id]["type"]


def test_load_frames_completeness(json_data, loader):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    assert len(scene.frames) == len(ground_truth["frames"])
    for frame_id, frame in scene.frames.items():

        assert str(frame_id) in ground_truth["frames"]
        assert frame.uid == frame_id


def test_load_frame_timestamps(json_data, loader):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    for frame_id, frame in scene.frames.items():
        assert str(frame.timestamp) == ground_truth["frames"][str(frame_id)]["frame_properties"]["timestamp"]


def test_load_frame_sensors(json_data, loader):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    for frame_id, frame in scene.frames.items():

        assert len(frame.sensors) == len(ground_truth["frames"][str(frame_id)]["frame_properties"]["streams"])
        for sensor_id, sensor in frame.sensors.items():

            assert sensor_id in ground_truth["frames"][str(frame_id)]["frame_properties"]["streams"]
            assert str(sensor.timestamp) == ground_truth["frames"][str(frame_id)]["frame_properties"]["streams"][sensor_id]["stream_properties"]["sync"]["timestamp"]
            assert str(sensor.uri) == ground_truth["frames"][str(frame_id)]["frame_properties"]["streams"][sensor_id]["uri"]


def test_load_frame_data(json_data, loader, annotation_compare_methods):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    for frame_id, frame in scene.frames.items():

        accumulative_frame_data = []
        for frame_data_type in ground_truth["frames"][str(frame_id)]["frame_properties"]["frame_data"].values():
            accumulative_frame_data.extend(frame_data_type)

        assert len(frame.frame_data) == len(accumulative_frame_data)

        for frame_data_type, frame_data in ground_truth["frames"][str(frame_id)]["frame_properties"]["frame_data"].items():
            for annotation in frame_data:
                annotation_compare_methods[frame_data_type](frame.frame_data[annotation["name"]], annotation)


def test_load_frame_annotations(json_data, loader, annotation_compare_methods):
    scene = loader.load(json_data["openlabel_v1_short"], validate=False)

    ground_truth = json_data["openlabel_v1_short"]["openlabel"]

    for frame_id, frame in scene.frames.items():

        assert len(frame.object_data) == len(ground_truth["frames"][str(frame_id)]["objects"])
        for object_id, object_data in ground_truth["frames"][str(frame_id)]["objects"].items():
            assert object_id in frame.object_data

            object_data = object_data["object_data"]

            accumulative_object_data = []
            for object_data_type in object_data.values():
                accumulative_object_data.extend(object_data_type)

            assert len(frame.object_data[object_id].annotations) == len(accumulative_object_data)

            for object_data_type, object_data in object_data.items():
                for annotation in object_data:
                    annotation_compare_methods[object_data_type](frame.annotations[annotation["uid"]], annotation)


def test_load_uri_vcd_incompatible(
    json_data, loader
):
    """Tests, whether an older annotation file, which is not usable for the VCD
    library is converted into a compatible one."""

    scene_ground_truth = loader.load(json_data["openlabel_v1_short"], validate=False)
    scene = loader.load(
        json_data["openlabel_v1_vcd_incompatible"],
        validate=False
    )

    # The UUIDs of the frame data have been generated and therefore do not match the ground truth.
    # They are set equal here.
    for frame_id in scene_ground_truth.frames:
        for frame_data in scene_ground_truth.frames[frame_id].frame_data:
            scene.frames[frame_id].frame_data[frame_data].uid = (
                scene_ground_truth.frames[frame_id].frame_data[frame_data].uid
            )

    assert scene == scene_ground_truth


# Tests the warnings and errors
def test_no_warnings(json_data, loader):
    loader.load(json_data["openlabel_v1_short"], validate=False)
    assert len(loader.warnings) == 0


def test_warning_uri_attribute(
    json_data, loader
):

    loader.load(
        json_data["openlabel_v1_vcd_incompatible"],
        validate=False
    )

    assert "attribute" in loader.warnings[0]
    assert "uri" in loader.warnings[0]


def test_stream_with_no_coordinate_system(json_data, loader):
    data = json_data["openlabel_v1_short"]

    del data["openlabel"]["coordinate_systems"]["ir_middle"]
    del data["openlabel"]["coordinate_systems"]["base"]["children"][
        data["openlabel"]["coordinate_systems"]["base"]["children"].index(
            "ir_middle"
        )
    ]
    with pytest.raises(raillabel.exceptions.MissingCoordinateSystemError):
        loader.load(data)


def test_warnings_sync(json_data, loader):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["frames"]["0"]["frame_properties"]["streams"]["non_existing_stream"] = {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321743.100000072"
            }
        }
    }

    loader.load(data, validate=False)
    assert len(loader.warnings) == 1

    # Tests for keywords in the warning that can help the user identify the source
    assert "frame" in loader.warnings[0]
    assert "0" in loader.warnings[0]
    assert "sync" in loader.warnings[0]
    assert "non_existing_stream" in loader.warnings[0]


def test_warnings_stream_sync_field(json_data, loader):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "rgb_middle"
    ]["stream_properties"]["stream_sync"] = data["openlabel"]["frames"]["0"][
        "frame_properties"
    ][
        "streams"
    ][
        "rgb_middle"
    ][
        "stream_properties"
    ][
        "sync"
    ]
    del data["openlabel"]["frames"]["0"]["frame_properties"]["streams"][
        "rgb_middle"
    ]["stream_properties"]["sync"]

    loader.load(data, validate=False)
    assert len(loader.warnings) == 1

    # Tests for keywords in the warning that can help the user identify the source
    assert "stream_sync" in loader.warnings[0]
    assert "deprecated" in loader.warnings[0].lower()
    assert "save()" in loader.warnings[0]


def test_warnings_ann_object(json_data, loader):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["frames"]["0"]["objects"][
        "affeaffe-0327-46ff-9c28-2506cfd6d934"
    ] = {"object_data": {}}

    loader.load(data, validate=False)
    assert len(loader.warnings) == 1

    assert not "affeaffe-0327-46ff-9c28-2506cfd6d934" in loader.scene.frames[0].object_data

    # Tests for keywords in the warning that can help the user identify the source
    assert "frame" in loader.warnings[0]
    assert "0" in loader.warnings[0]
    assert "object" in loader.warnings[0]
    assert "affeaffe-0327-46ff-9c28-2506cfd6d934" in loader.warnings[0]


def test_warnings_wrong_annotation_cs(json_data, loader):
    data = json_data["openlabel_v1_short"]

    data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["bbox"][0]["coordinate_system"] = "non_existent_sensor"

    loader.load(data, validate=False)
    assert len(loader.warnings) == 1

    assert (
        loader.scene.frames[0]
        .object_data["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
        .bboxs["78f0ad89-2750-4a30-9d66-44c9da73a714"]
        .sensor
        is None
    )

    # Tests for keywords in the warning that can help the user identify the source
    assert "annotation" in loader.warnings[0]
    assert "78f0ad89-2750-4a30-9d66-44c9da73a714" in loader.warnings[0]
    assert "sensor" in loader.warnings[0]
    assert "non_existent_sensor" in loader.warnings[0]


def test_identify_of_references(json_data, loader):
    data = json_data["openlabel_v1_short"]

    scene = loader.load(data, validate=False)

    for frame in scene.frames.values():

        for sensor_reference in frame.sensors.values():
            assert sensor_reference.sensor is scene.sensors[sensor_reference.sensor.uid]

        for frame_data in frame.frame_data.values():
            assert frame_data.sensor is scene.sensors[frame_data.sensor.uid]

        for annotation in frame.annotations.values():
            assert annotation.sensor is scene.sensors[annotation.sensor.uid]


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Module containing all relevant OpenLABEL format classes."""

from .bbox import Bbox
from .cuboid import Cuboid
from .frame import Frame
from .frame_interval import FrameInterval
from .intrinsics_pinhole import IntrinsicsPinhole
from .intrinsics_radar import IntrinsicsRadar
from .metadata import Metadata
from .num import Num
from .object import Object
from .object_data import ObjectData
from .point2d import Point2d
from .point3d import Point3d
from .poly2d import Poly2d
from .poly3d import Poly3d
from .quaternion import Quaternion
from .scene import Scene
from .seg3d import Seg3d
from .sensor import Sensor, SensorType
from .sensor_reference import SensorReference
from .size2d import Size2d
from .size3d import Size3d
from .transform import Transform



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
import uuid
from dataclasses import dataclass, field

from .._util._warning import _warning
from ._annotation import _Annotation
from .bbox import Bbox
from .cuboid import Cuboid
from .object import Object
from .poly2d import Poly2d
from .poly3d import Poly3d
from .seg3d import Seg3d
from .sensor import Sensor


@dataclass
class ObjectData:
    """Annotations associated with a specific object in a frame.

    Parameters
    ----------
    object: raillabel.format.Object
        A reference to the object this ObjectData belongs to.
    annotations: dict, optional
        Dictionary of all annotations representing this object in this frame.

    Properties (read-only)
    ----------------------
    bboxs: dict of raillabel.format.Bbox
        Dictionary of all bounding boxes representing this object in this frame.
    cuboids: dict of raillabel.format.Cuboid
        Dictionary of all cuboids representing this object in this frame.
    poly2ds: dict of raillabel.format.Poly2d
        Dictionary of all 2d polylines representing this object in this frame.
    poly3ds: dict of raillabel.format.Poly3d
        Dictionary of all 3d polylines representing this object in this frame.
    seg3ds: dict of raillabel.format.Seg3d
        Dictionary of all 3d segmentations representing this object in this frame.
    """

    object: Object
    annotations: t.Dict[str, t.Type[_Annotation]] = field(default_factory=dict)

    @property
    def bboxs(self) -> t.Dict[str, Bbox]:
        """Return dictionary of all bounding boxes."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Bbox)}

    @property
    def cuboids(self) -> t.Dict[str, Cuboid]:
        """Return dictionary of all cuboids."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Cuboid)}

    @property
    def poly2ds(self) -> t.Dict[str, Poly2d]:
        """Return dictionary of all 2d poly lines."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Poly2d)}

    @property
    def poly3ds(self) -> t.Dict[str, Bbox]:
        """Return dictionary of all 3d poly lines."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Poly3d)}

    @property
    def seg3ds(self) -> t.Dict[str, Seg3d]:
        """Return dictionary of all 3d segmentations."""
        return {k: v for k, v in self.annotations.items() if isinstance(v, Seg3d)}

    @classmethod
    def fromdict(
        cls,
        uid: str,
        data_dict: dict,
        objects: t.Dict[str, Object],
        sensors: t.Dict[str, Sensor],
        annotation_classes: dict,
    ) -> "ObjectData":
        """Generate a ObjectData object from a dict.

        Parameters
        ----------
        uid: str
            Unique identifier of the object.
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        objects: dict
            Dictionary of all objects in the scene.
        sensors: dict
            Dictionary of all sensors in the scene.
        annotation_classes: dict
            Dictionary conaining all of the annotation classes as values
            with the OpenLABEL identifiers as keys.

        Returns
        -------
        object_data: raillabel.format.ObjectData
            Converted ObjectData object.
        """

        object_data = ObjectData(object=objects[uid])

        for ann_type in data_dict:

            if ann_type not in annotation_classes:
                _warning(
                    f"Annotation type {ann_type} is currently not supported. Supported "
                    + "annotation types: "
                    + str(list(annotation_classes.keys()))
                )
                continue

            for ann_raw in data_dict[ann_type]:

                ann_raw = cls._fix_deprecated_annotation_name(ann_raw, ann_type, objects[uid].type)

                if ann_raw["uid"] in object_data.annotations:
                    _warning(
                        f"Annotation '{ann_raw['uid']}' is contained more than one "
                        + "time. A new UID is beeing assigned."
                    )
                    ann_raw["uid"] = str(uuid.uuid4())

                object_data.annotations[ann_raw["uid"]] = annotation_classes[ann_type].fromdict(
                    ann_raw,
                    sensors,
                )

        return object_data

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {"object_data": {}}

        if self.bboxs != {}:
            dict_repr["object_data"]["bbox"] = []

            for bbox in self.bboxs.values():
                dict_repr["object_data"]["bbox"].append(bbox.asdict())

        if self.poly2ds != {}:
            dict_repr["object_data"]["poly2d"] = []

            for poly2d in self.poly2ds.values():
                dict_repr["object_data"]["poly2d"].append(poly2d.asdict())

        if self.poly3ds != {}:
            dict_repr["object_data"]["poly3d"] = []

            for poly3d in self.poly3ds.values():
                dict_repr["object_data"]["poly3d"].append(poly3d.asdict())

        if self.cuboids != {}:
            dict_repr["object_data"]["cuboid"] = []

            for cuboid in self.cuboids.values():
                dict_repr["object_data"]["cuboid"].append(cuboid.asdict())

        if self.seg3ds != {}:
            dict_repr["object_data"]["vec"] = []

            for seg3d in self.seg3ds.values():
                dict_repr["object_data"]["vec"].append(seg3d.asdict())

        return dict_repr

    @classmethod
    def _fix_deprecated_annotation_name(cls, ann_raw: dict, ann_type: str, obj_type: str) -> dict:

        if "uid" not in ann_raw:
            try:
                ann_raw["uid"] = str(uuid.UUID(ann_raw["name"]))
            except ValueError:
                ann_raw["uid"] = str(uuid.uuid4())

        ann_raw["name"] = f"{ann_raw['coordinate_system']}__{ann_type}__{obj_type}"

        return ann_raw



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass

from .coordinate_system import CoordinateSystem
from .frame import Frame
from .metadata import Metadata


@dataclass
class Scene:
    """The root Understand.Ai class, which contains all data.

    Parameters
    ----------
    metadata: raillabel._understand_ai_t4_format.Metadata
        Container for metadata information about the scene itself.
    coordinate_systems: dict[str, raillabel._understand_ai_t4_format.CoordinateSystem]
        Global information for sensors regarding calibration.
    frames: dict[int, raillabel._understand_ai_t4_format.Frame]
    """

    metadata: Metadata
    coordinate_systems: t.Dict[str, CoordinateSystem]
    frames: t.Dict[int, Frame]

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Scene":
        """Generate a Scene from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Scene
            Converted scene.
        """

        return Scene(
            metadata=Metadata.fromdict(data_dict["metadata"]),
            coordinate_systems=cls._coordinate_systems_fromdict(data_dict["coordinateSystems"]),
            frames=cls._frames_fromdict(data_dict["frames"]),
        )

    def to_raillabel(self) -> t.Tuple[dict, dict]:
        """Convert to a raillabel compatible dict.

        Returns
        -------
        dict:
            Dictionary of the raillabel scene.
        """
        return {
            "openlabel": {
                "metadata": self.metadata.to_raillabel(),
                "streams": self._streams_to_raillabel(),
                "coordinate_systems": self._coordinate_systems_to_raillabel(),
                "objects": self._objects_to_raillabel(),
                "frames": {str(frame.id): frame.to_raillabel() for frame in self.frames.values()},
            }
        }

    @classmethod
    def _coordinate_systems_fromdict(cls, data_dict: t.List[dict]) -> t.Dict[str, CoordinateSystem]:
        coordinate_systems = {}
        for cs in data_dict:
            coordinate_systems[cs["coordinate_system_id"]] = CoordinateSystem.fromdict(cs)

        return coordinate_systems

    @classmethod
    def _frames_fromdict(cls, data_dict: t.List[dict]) -> t.Dict[str, Frame]:
        frames = {}
        for frame in data_dict:
            frames[frame["frameId"]] = Frame.fromdict(frame)

        return frames

    def _streams_to_raillabel(self) -> dict:
        return {cs.translated_uid: cs.to_raillabel()[1] for cs in self.coordinate_systems.values()}

    def _coordinate_systems_to_raillabel(self) -> dict:
        coordinate_systems = {
            cs.translated_uid: cs.to_raillabel()[0] for cs in self.coordinate_systems.values()
        }

        coordinate_systems["base"] = {
            "type": "local",
            "parent": "",
            "children": list(coordinate_systems.keys()),
        }

        return coordinate_systems

    def _objects_to_raillabel(self) -> dict:

        object_dicts = self._collect_all_translated_objects()

        object_name_counter = {}
        objects = {}
        for object_id, object_class in object_dicts.items():

            if object_class not in object_name_counter:
                object_name_counter[object_class] = 0

            objects[object_id] = {
                "name": f"{object_class}_{str(object_name_counter[object_class]).rjust(4, '0')}",
                "type": object_class,
            }
            object_name_counter[object_class] += 1

        return objects

    def _collect_all_translated_objects(self) -> dict:
        object_dicts = {}
        for frame in self.frames.values():
            object_dicts.update(frame.translated_objects)

        return object_dicts



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass
from uuid import UUID

from ._translation import translate_class_id, translate_sensor_id
from .sensor_reference import SensorReference


@dataclass
class _Annotation(ABC):

    id: UUID
    object_id: UUID
    class_name: str
    attributes: dict
    sensor: SensorReference

    @property
    @abstractproperty
    def OPENLABEL_ID(self) -> t.List[str]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def fromdict(cls, data_dict: t.Dict) -> t.Type["_Annotation"]:
        raise NotImplementedError

    def to_raillabel(self) -> t.Tuple[dict, str, str, dict]:
        """Convert to a raillabel compatible dict.

        Returns
        -------
        annotation: dict
            Dictionary valid for the raillabel schema.
        object_id: str
            Friendly identifier of the object this annotation belongs to.
        class_name: str
            Friendly identifier of the class the annotated object belongs to.
        sensor_reference: dict
            Dictionary of the sensor reference.
        """

        return (
            {
                "name": str(self.id),
                "val": self._val_to_raillabel(),
                "coordinate_system": translate_sensor_id(self.sensor.type),
                "attributes": self._attributes_to_raillabel(),
            },
            str(self.object_id),
            translate_class_id(self.class_name),
            self.sensor.to_raillabel()[1],
        )

    def _attributes_to_raillabel(self) -> dict:

        attributes = {}

        for attr_name, attr_value in self.attributes.items():

            if type(attr_value) == str:
                attr_type = "text"

            elif type(attr_value) == bool:
                attr_type = "boolean"

            elif type(attr_value) in [float, int]:
                attr_type = "num"

            elif type(attr_value) in [list, tuple]:
                attr_type = "vec"

            else:
                raise TypeError(
                    f"Attribute type {attr_value.__class__.__name__} of {attr_value} is not "
                    + "supported. Supported types are str, float, int, bool, list, tuple."
                )

            if attr_type not in attributes:
                attributes[attr_type] = []

            attributes[attr_type].append({"name": attr_name, "val": attr_value})

        return attributes



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import pickle

from . import _filter_classes, format


def filter(scene: format.Scene, **kwargs) -> format.Scene:
    """Return a copy of the scene with the annotations filtered.

    Parameters
    ----------
    scene: raillabel.Scene
        Scene, which should be copied and filtered.
    include_object_types: str or list of str, optional
        List of class/type names that should be included in the filtered scene. If set, no
        other classes/types will be copied. Mutually exclusive with exclude_object_types.
    exclude_object_types: str or list of str, optional
        List of class/type names that should be excluded in the filtered scene. If set, all
        other classes/types will be copied. Mutually exclusive with include_object_types.
    include_annotation_types: str or list of str, optional
        List of annotation types (i.e. bboxs, cuboids, poly2ds, seg3ds) that should be included
        in the filtered scene. If set, no other annotation types will be copied. Mutually
        exclusive with exclude_annotation_types.
    exclude_annotation_types: str or list of str, optional
        List of annotation types (i.e. bboxs, cuboids, poly2ds, seg3ds) that should be excluded
        in the filtered scene. If set, all other annotation types will be copied. Mutually
        exclusive with include_annotation_types.
    include_annotation_ids: str or list of str, optional
        List of annotation UIDs that should be included in the filtered scene. If set, no other
        annotation UIDs will be copied. Mutually exclusive with exclude_annotation_ids.
    exclude_annotation_ids: str or list of str, optional
        List of annotation UIDs that should be excluded in the filtered scene. If set, all
        other annotation UIDs will be copied. Mutually exclusive with include_annotation_ids.
    include_object_ids: str or list of str, optional
        List of object UIDs that should be included in the filtered scene. If set, no other
        objects will be copied. Mutually exclusive with exclude_object_ids.
    exclude_object_ids: str or list of str, optional
        List of object UIDs that should be excluded in the filtered scene. If set, all other
        objects will be copied. Mutually exclusive with include_object_ids.
    include_sensors: str or list of str
        List of sensors that should be included in the filtered scene. If set, no other
        sensors will be copied. Mutually exclusive with exclude_sensors.
    exclude_sensors: str or list of str, optional
        List of sensors that should be excluded in the filtered scene. If set, all other
        sensors will be copied. Mutually exclusive with include_sensors.
    include_attributes: dict, optional
        Dict of attributes that should be included in the filtered scene. Dict keys are the
        attribute names, values are the specific values that should be included. If the
        value is set so None, all annotations with the attribute are included regardless of
        value. Mutually exclusive with exclude_attributes.
    exclude_attributes: dict, optional
        Dict of attributes that should be excluded in the filtered scene. Dict keys are the
        attribute names, values are the specific values that should be excluded. If the value
        is set so None, all annotations with the attribute are excluded regardless of value.
        Mutually exclusive with include_attributes.
    include_frames: int or list of int, optional
        List of frame UIDs that should be included in the filtered scene. If set, no other
        frames will be copied. Mutually exclusive with exclude_frames.
    exclude_frames: int or list of int, optional
        List of frame UIDs that should be excluded in the filtered scene. If set, all other
        frames will be copied. Mutually exclusive with include_frames.
    start_frame: int, optional
        Frame at which the filtered scene should start. Mutually exclusive with s
        tart_timestamp.
    end_frame: int, optional
        Frame at which the filtered scene should end (inclusive). Mutually exclusive with
        end_timestamp.
    start_timestamp: decimal.Decimal, optional
        Unix timestamp at which the filtered scene should start (inclusive). Mutually exclusive
        with start_frame.
    end_timestamp: decimal.Decimal, optional
        Unix timestamp at which the filtered scene should end (inclusive). Mutually exclusive
        with end_frame.

    Raises
    ------
    ValueError
        if two mutually exclusive parameters are set.
    TypeError
        if an unexpected keyword argument has been set.
    """

    # Collects the filters
    filters = []
    supported_kwargs = []
    for cls in _filter_classes.__dict__.values():
        if (
            isinstance(cls, type)
            and issubclass(cls, _filter_classes._FilterABC)
            and cls != _filter_classes._FilterABC
        ):
            filters.append(cls(kwargs))
            supported_kwargs.extend(cls.PARAMETERS)

    frame_filters = [f for f in filters if "frame" in f.LEVELS]
    frame_data_filters = [f for f in filters if "frame_data" in f.LEVELS]
    object_filters = [f for f in filters if "object" in f.LEVELS]
    annotation_filters = [f for f in filters if "annotation" in f.LEVELS]

    # Raises an exception if an unexpected kwarg is receifed
    for arg in kwargs:
        if arg not in supported_kwargs:
            raise TypeError(
                f"filter() got an unexpected keyword argument '{arg}'. Supported keyword "
                + f"arguments: {sorted(supported_kwargs)}"
            )

    filtered_scene = _copy(scene)

    used_sensors = set()
    used_objects = set()

    for frame_id, frame in scene.frames.items():

        if not _passes_filters(frame, frame_filters):
            del filtered_scene.frames[frame_id]
            continue

        for frame_data_id, frame_data in frame.frame_data.items():

            if _passes_filters(frame_data, frame_data_filters):
                used_sensors.add(frame_data.sensor.uid)

            else:
                del filtered_scene.frames[frame_id].frame_data[frame_data_id]

        for object_id, object_data in frame.object_data.items():

            if not _passes_filters(scene.objects[object_id], object_filters):
                continue

            for annotation_id, annotation in object_data.annotations.items():

                if _passes_filters(annotation, annotation_filters):
                    used_objects.add(object_id)
                    used_sensors.add(annotation.sensor.uid)

                else:
                    del (
                        filtered_scene.frames[frame_id]
                        .object_data[object_id]
                        .annotations[annotation_id]
                    )

    # Clears out any unused sensors, objects and references

    for sensor_id in scene.sensors:
        if sensor_id not in used_sensors:
            del filtered_scene.sensors[sensor_id]

    for object_id in scene.objects:
        if object_id not in used_objects:
            del filtered_scene.objects[object_id]

    for frame_id in filtered_scene.frames:

        for sensor_id in scene.frames[frame_id].sensors:
            if sensor_id not in used_sensors:
                del filtered_scene.frames[frame_id].sensors[sensor_id]

        for object_id in scene.frames[frame_id].object_data:

            if (
                object_id not in used_objects
                or len(filtered_scene.frames[frame_id].object_data[object_id].annotations) == 0
            ):
                del filtered_scene.frames[frame_id].object_data[object_id]

    return filtered_scene


def _passes_filters(data, filters):
    for f in filters:
        if not f.passes_filter(data):
            return False

    return True


def _copy(object):
    return pickle.loads(pickle.dumps(object, -1))



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os

import jsonschema
import pytest


def test_metaschema_validation(json_data):
    assert jsonschema.validate(
        json_data["understand_ai_t4_schema"],
        json_data["metaschema"]
    ) is None


def test_sample_data_validation_subschema(json_data):
    assert jsonschema.validate(
        json_data["understand_ai_real_life"],
        json_data["understand_ai_t4_schema"]
    ) is None

    assert jsonschema.validate(
        json_data["understand_ai_t4_short"],
        json_data["understand_ai_t4_schema"]
    ) is None


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB Netz AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
from pathlib import Path

from . import exceptions
from .format import Scene
from .validate import validate as validate_func


def save(scene: Scene, path: str, prettify_json: bool = False, validate: bool = False):
    """Save a raillabel.Scene in a JSON file.

    Parameters
    ----------
    scene: raillabel.Scene
        Scene, which should be saved.
    path: str
        Path to the file location, that should be used for saving.
    save_path: str
        Path to the JSON file.
    prettify_json: bool, optional
        If true, the JSON is saved with linebreaks and indents. This increases readibility but
        also the file size. Default is False.
    validate: bool, optional
        If True, the annotation data is validated via the OpenLabel schema. This is highly
        recommended, as not validating the data may lead to Errors during loading or while handling
        the scene. However, validating may increase the loading time. Default is False.

    Raises
    ------
    SchemaError
        If the data does not validate via the OpenLabel schema.
    FileNotFoundError
        If the save_path does not point to an accessible file.
    """

    path = Path(path)

    # Converts the data stored in the scene into a dictionary
    data = scene.asdict()

    # Validates the data
    if validate:
        is_data_valid, err_msgs = validate_func(data)
        if not is_data_valid:
            schema_err_msg = (
                "The data could not be saved, because it does not validate "
                "against the OpenLabel schema:"
            )

            for err_msg in err_msgs:
                schema_err_msg += "\n - " + err_msg

            raise exceptions.SchemaError(schema_err_msg)

    # Saves the data
    with path.open("w") as save_file:

        if prettify_json:
            json.dump(data, save_file, indent=4)
        else:
            json.dump(data, save_file)
